import os
import pathlib
import json
import pickle
import requests
import numpy as np
import pandas as pd
import scanpy as sc
from anndata import AnnData
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LogisticRegression
from typing import Optional, Union
from scipy.special import expit
from scipy.sparse import spmatrix
from datetime import datetime
import logging
from dance.transforms.preprocess import prepare_data_celltypist, to_array_celltypist, to_vector_celltypist, downsample_adata, get_sample_csv_celltypist, get_sample_data_celltypist, LRClassifier_celltypist, SGDClassifier_celltypist
from typing import Optional, Union
from matplotlib import pyplot as plt
import warnings
import ipdb


logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)
set_level = logger.setLevel
info = logger.info
warn = logger.warning
error = logger.error
debug = logger.debug

celltypist_path = os.getenv('CELLTYPIST_FOLDER', default = os.path.join(str(pathlib.Path.home()), '.celltypist'))
pathlib.Path(celltypist_path).mkdir(parents=True, exist_ok=True)
data_path = os.path.join(celltypist_path, "data")
models_path = os.path.join(data_path, "models")
pathlib.Path(models_path).mkdir(parents=True, exist_ok=True)

_samples_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "samples")



def _collapse_mean(arr: np.ndarray) -> Union[float, np.ndarray]:
    """
    For internal use. Average 1D array, or 2D array by row.
    """
    return np.mean(arr, axis = -1)

def _collapse_random(arr: np.ndarray) -> Union[float, np.ndarray]:
    """
    For internal use. Choose a random number from 1D array, or a random column from 2D array.
    """
    return np.random.choice(arr, 1)[0] if arr.ndim == 1 else arr[:, np.random.choice(arr.shape[1], 1)[0]]

class Model():
    """
    Class that wraps the logistic Classifier and the StandardScaler.
    
    Parameters
    ----------
    clf
        A logistic Classifier incorporated in the loaded model.
    scaler
        A StandardScaler incorporated in the loaded model.
    description
        Description of the model as a dictionary.
        
    Attributes
    ----------
    classifier
        The logistic Classifier incorporated in the loaded model.
    scaler
        The StandardScaler incorporated in the loaded model.
    description
        Description of the loaded model.
    """
    def __init__(self, clf, scaler, description):
        self.classifier = clf
        self.scaler = scaler
        self.description = description

    @staticmethod
    def load(self, model: Optional[str] = None):
        """
        Load the desired model.
        
        Parameters
        ----------
        model: Optional[str]
            Model name specifying the model you want to load. Default to `'Immune_All_Low.pkl'` if not provided.
            To see all available models and their descriptions, use :func:`~celltypist.models.models_description`.
            
        Returns
        ----------
        :class:`~celltypist.models.Model`
            A :class:`~celltypist.models.Model` object.
        """
        if not model:
            model = get_default_model()
        if model in get_all_models():
            model = get_model_path(model)
        if not os.path.isfile(model):
            raise FileNotFoundError(f"?? No such file: {model}")
        with open(model, "rb") as fh:
            try:
                pkl_obj = pickle.load(fh)
                return Model(pkl_obj['Model'], pkl_obj['Scaler_'], pkl_obj['description'])
            except Exception as exception:
                raise Exception(f"?? Invalid model: {model}. {exception}")

    @property
    def cell_types(self) -> np.ndarray:
        """Get cell types included in the model."""
        return self.classifier.classes_

    @property
    def features(self) -> np.ndarray:
        """Get genes included in the model."""
        return self.classifier.features

    def __repr__(self):
        """
        General the description
        """
        base = f"CellTypist model with {len(self.cell_types)} cell types and {len(self.features)} features"
        for x in ['date', 'details', 'source', 'version']:
            if self.description[x] != '':
                base += f"\n    {x}: {self.description[x]}"
        if len(self.cell_types) == 2:
            base += f"\n    cell types: {self.cell_types[0]}, {self.cell_types[1]}\n    features: {self.features[0]}, {self.features[1]}, ..., {self.features[-1]}"
        else:
            base += f"\n    cell types: {self.cell_types[0]}, {self.cell_types[1]}, ..., {self.cell_types[-1]}\n    features: {self.features[0]}, {self.features[1]}, ..., {self.features[-1]}"
        return base

    def predict_labels_and_prob(self, indata, mode: str = 'best match', p_thres: float = 0.5) -> tuple:
        """
        Get the decision matrix, probability matrix, and predicted cell types for the input data.
        
        Parameters
        ----------
        indata
            The input array-like object used as a query.
        mode: str
            The way cell prediction is performed.
            For each query cell, the default (`'best match'`) is to choose the cell type with the largest score/probability as the final prediction.
            Setting to `'prob match'` will enable a multi-label classification, which assigns 0 (i.e., unassigned), 1, or >=2 cell type labels to each query cell.
            (Default: `'best match'`)
        p_thres: float
            Probability threshold for the multi-label classification. Ignored if `mode` is `'best match'`.
            (Default: 0.5)
            
        Returns
        ----------
        tuple
            A tuple of decision score matrix, raw probability matrix, and predicted cell type labels.
        """
        scores = self.classifier.decision_function(indata)
        if len(self.cell_types) == 2:
            scores = np.column_stack([-scores, scores])
        probs = expit(scores)
        if mode == 'best match':
            return scores, probs, self.classifier.classes_[scores.argmax(axis=1)]
        elif mode == 'prob match':
            flags = probs > p_thres
            labs = np.array(['|'.join(self.classifier.classes_[np.where(x)[0]]) for x in flags])
            labs[labs == ''] = 'Unassigned'
            return scores, probs, labs
        else:
            raise ValueError(f"?? Unrecognized `mode` value, should be one of `'best match'` or `'prob match'`")

    def write(self, file: str) -> None:
        """Write out the model."""
        obj = dict(Model = self.classifier, Scaler_ = self.scaler, description = self.description)
        file = os.path.splitext(file)[0] + '.pkl'
        with open(file, 'wb') as output:
            pickle.dump(obj, output)

    def extract_top_markers(self, cell_type: str, top_n: int = 10, only_positive: bool = True) -> np.ndarray:
        """
        Extract the top driving genes for a given cell type.
        
        Parameters
        ----------
        cell_type: str
            The cell type to extract markers for.
        top_n: int
            Number of markers to extract for a given cell type.
            (Default: 10)
        only_positive: bool
            Whether to extract positive markers only. Set to `False` to include negative markers as well.
            (Default: `True`)
            
        Returns
        ----------
        :class:`~numpy.ndarray`
            A list of marker genes for the query cell type.
        """
        if cell_type not in self.cell_types:
            raise ValueError(f"?? '{cell_type}' is not found. Please provide a valid cell type name")
        if len(self.cell_types) == 2:
            coef_vector = self.classifier.coef_[0] if cell_type == self.cell_types[1] else -self.classifier.coef_[0]
        else:
            coef_vector = self.classifier.coef_[self.cell_types == cell_type][0]
        if not only_positive:
            coef_vector = np.abs(coef_vector)
        return self.features[np.argsort(-coef_vector)][:top_n]

    def convert(self, map_file: Optional[str] = None, sep: str = ',', convert_from: Optional[int] = None, convert_to: Optional[int] = None, unique_only: bool = True, collapse: str = 'average', random_state: int = 0) -> None:
        """
        Convert the model of one species to another species by mapping orthologous genes.
        
        Parameters
        ----------
        map_file: Optional[str]
            A two-column gene mapping file between two species.
            Default to a human-mouse (mouse-human) conversion using the built-in mapping file provided by CellTypist.
        sep: str
            Delimiter of the mapping file. Default to comma (i.e., a csv file is by default expected from the user if provided).
        convert_from: Optional[int]
            Column index (0 or 1) of the mapping file corresponding to the species converted from.
            Default to an automatic detection.
        convert_to: Optional[int]
            Column index (0 or 1) of the mapping file corresponding to the species converted to.
            Default to an automatic detection.
        unique_only: bool
            Whether to leverage only 1:1 orthologs between the two species.
            (Default: `True`)
        collapse: str
            The way 1:N orthologs are handled. Possible values are `'average'` which averages the classifier weights and `'random'` which randomly chooses one gene's weights from all its orthologs.
            This argument is ignored if `unique_only = True`.
            (Default: `'average'`)
        random_state: int
            Random seed for reproducibility. This argument is only relevant if `unique_only = False` and `collapse = 'random'`.
            
        Returns
        ----------
        None
            The original model is modified by converting to the other species.
        """
        map_file = get_sample_data_celltypist('Ensembl105_Human2Mouse_Genes.csv') if map_file is None else map_file
        if not os.path.isfile(map_file):
            raise FileNotFoundError(f"?? No such file: {map_file}")
        #with and without headers are both ok -> real headers become fake genes and are removed afterwards
        map_content = pd.read_csv(map_file, sep = sep, header = None)
        map_content.dropna(axis = 0, inplace = True)
        map_content.drop_duplicates(inplace = True)
        #From & To detection
        if (convert_from is None) and (convert_to is None):
            column1_overlap = map_content[0].isin(self.features).sum()
            column2_overlap = map_content[1].isin(self.features).sum()
            convert_from = 0 if column1_overlap > column2_overlap else 1
            convert_to = 1 - convert_from
        elif convert_from is None:
            if convert_to not in [0, 1]:
                raise ValueError(f"?? `convert_to` should be either 0 or 1")
            convert_from = 1 - convert_to
        elif convert_to is None:
            if convert_from not in [0, 1]:
                raise ValueError(f"?? `convert_from` should be either 0 or 1")
            convert_to = 1 - convert_from
        else:
            if {convert_from, convert_to} != {0, 1}:
                raise ValueError(f"?? `convert_from` and `convert_to` should be 0 (or 1) and 1 (or 0)")
        #filter
        map_content = map_content[map_content[convert_from].isin(self.features)]
        if unique_only:
            map_content.drop_duplicates([0],inplace=True)
            map_content.drop_duplicates([1],inplace=True)
        map_content['index_from'] = pd.DataFrame(self.features, columns=['features']).reset_index().set_index('features').loc[map_content[convert_from], 'index'].values
        #main
        logger.info(f"?? Number of genes in the original model: {len(self.features)}")
        features_to = map_content[convert_to].values if unique_only else np.unique(map_content[convert_to])
        if unique_only:
            index_from = map_content['index_from'].values
            self.classifier.coef_ = self.classifier.coef_[:, index_from]
            self.scaler.mean_ = self.scaler.mean_[index_from]
            self.scaler.var_ = self.scaler.var_[index_from]
            self.scaler.scale_ = self.scaler.scale_[index_from]
        else:
            if collapse not in ['average', 'random']:
                raise ValueError(f"?? Unrecognized `collapse` value, should be one of `'average'` or `'random'`")
            if collapse == 'random':
                np.random.seed(random_state)
            collapse_func = _collapse_mean if collapse == 'average' else _collapse_random
            coef_to = []
            mean_to = []
            var_to = []
            scale_to = []
            for feature_to in features_to:
                index_from = map_content[map_content[convert_to] == feature_to].index_from.values
                if len(index_from) == 1:
                    coef_to.append(self.classifier.coef_[:, index_from[0]])
                    mean_to.append(self.scaler.mean_[index_from[0]])
                    var_to.append(self.scaler.var_[index_from[0]])
                    scale_to.append(self.scaler.scale_[index_from[0]])
                else:
                    coef_to.append(collapse_func(self.classifier.coef_[:, index_from]))
                    mean_to.append(collapse_func(self.scaler.mean_[index_from]))
                    var_to.append(collapse_func(self.scaler.var_[index_from]))
                    scale_to.append(collapse_func(self.scaler.scale_[index_from]))
            self.classifier.coef_ = np.column_stack(coef_to)
            self.scaler.mean_ = np.array(mean_to)
            self.scaler.var_ = np.array(var_to)
            self.scaler.scale_ = np.array(scale_to)
        self.classifier.n_features_in_ = len(features_to)
        self.classifier.features = features_to
        self.scaler.n_features_in_ = len(features_to)
        logger.info(f"? Conversion done! Number of genes in the converted model: {len(features_to)}")

def get_model_path(file: str) -> str:
    """
    Get the full path to a file in the `models` folder.
    
    Parameters
    ----------
    file: str
        File name as a string.
        To see all available models and their descriptions, use :func:`~celltypist.models.models_description`.
        
    Returns
    ----------
    str
        A string of the full path to the desired file.
    """
    return os.path.join(models_path, f"{file}")


def get_default_model() -> str:
    """
    Get the default model name.
    
    Returns
    ----------
    str
        A string showing the default model name (should be `'Immune_All_Low.pkl'`).
    """
    models_json = get_models_index()
    default_model = [m["filename"] for m in models_json["models"] if ("default" in m and m["default"])]
    if not default_model:
        first_model = models_json["models"][0]["filename"]
        logger.warn(f"?? No model marked as 'default', using {first_model}")
        return first_model
    if len(default_model) > 1:
        logger.warn(f"?? More than one model marked as 'default', using {default_model[0]}")
    return default_model[0]


def get_all_models() -> list:
    """
    Get a list of all the available models.
    Returns
    ----------
    list
        A list of available models.
    """
    download_if_required()
    available_models = []
    for model_filename in os.listdir(models_path):
        if model_filename.endswith(".pkl"):
            model_name = os.path.basename(model_filename)
            available_models.append(model_name)
    return available_models


def download_if_required() -> None:
    """Download models if there are none present in the `models` directory."""
    if len([m for m in os.listdir(models_path) if m.endswith(".pkl")]) == 0:
        logger.info(f"?? No available models. Downloading...")
        download_models()


def get_models_index(force_update: bool=False) -> dict:
    """
    Get the model json object containing the model list.
    
    Parameters
    ----------
    force_update: bool
        If set to `True`, will download the latest model json file from the remote.
        (Default: `False`)
        
    Returns
    ----------
    dict: dict
        A dict object converted from the model json file.
    """
    models_json_path = get_model_path("models.json")
    if not os.path.exists(models_json_path) or force_update:
        download_model_index()
    with open(models_json_path) as f:
        return json.load(f)


def download_model_index(only_model: bool = True) -> None:
    """
    Download the `models.json` file from the remote server.
    
    Parameters
    ----------
    only_model: bool
        If set to `False`, will also download the models in addition to the json file.
        (Default: `True`)
        
    Returns
    ----------
    No return
    """
    url = 'https://celltypist.cog.sanger.ac.uk/models/models.json'
    logger.info(f"?? Retrieving model list from server {url}")
    with open(get_model_path("models.json"), "wb") as f:
        f.write(requests.get(url).content)
    model_count = len(requests.get(url).json()["models"])
    logger.info(f"?? Total models in list: {model_count}")
    if not only_model:
        download_models()

def download_models(force_update: bool=False, model: Optional[Union[str, list, tuple]] = None) -> None:
    """
    Download all the available or selected models.
    
    Parameters
    ----------
    force_update: bool
        Whether to fetch a latest JSON index for downloading all available or selected models.
        Set to `True` if you want to parallel the latest celltypist model releases.
        (Default: `False`)
    model: Optional[Union[str, list, tuple]]
        Specific model(s) to download. By default, all available models are downloaded.
        Set to a specific model name or a list of model names to only download a subset of models.
        For example, set to `["ModelA.pkl", "ModelB.pkl"]` to only download ModelA and ModelB.
        To check all available models, use :func:`~celltypist.models.models_description`.
    """
    models_json = get_models_index(force_update)
    logger.info(f"?? Storing models in {models_path}")
    if model is not None:
        model_list = {model} if isinstance(model, str) else set(model)
        models_json["models"] = [m for m in models_json["models"] if m["filename"] in model_list]
        provided_no = len(model_list)
        filtered_no = len(models_json["models"])
        if filtered_no == 0:
            raise ValueError(f"?? No models match the celltypist model repertoire. Please provide valid model names")
        elif provided_no == filtered_no:
            logger.info(f"?? Total models to download: {provided_no}")
        else:
            ignored_models = model_list.difference({m["filename"] for m in models_json["models"]})
            logger.warn(f"?? Total models to download: {filtered_no}. {len(ignored_models)} not available: {ignored_models}")
    model_count = len(models_json["models"])
    for idx,model in enumerate(models_json["models"]):
        model_path = get_model_path(model["filename"])
        if os.path.exists(model_path) and not force_update:
            logger.info(f"? Skipping [{idx+1}/{model_count}]: {model['filename']} (file exists)")
            continue
        logger.info(f"?? Downloading model [{idx+1}/{model_count}]: {model['filename']}")
        try:
            with open(model_path, "wb") as f:
                f.write(requests.get(model["url"]).content)
        except Exception as exception:
            logger.error(f"?? {model['filename']} failed {exception}")


def models_description(on_the_fly: bool=False) -> pd.DataFrame:
    """
    Get the descriptions of all available models.
    
    Parameters
    ----------
    on_the_fly: bool
        Whether to fetch the model information from downloaded model files.
        If set to `True`, will fetch the information by loading downloaded models.
        Default to fetching the information for all available models from the JSON file.
        (Default: `False`)
        
    Returns
    ----------
    :class:`~pandas.DataFrame`
        A :class:`~pandas.DataFrame` object with model descriptions.
    """
    logger.info(f"?? Detailed model information can be found at `https://www.celltypist.org/models`")
    if on_the_fly:
        filenames = get_all_models()
        descriptions = [Model.load(filename).description['details'] for filename in filenames]
    else:
        models_json = get_models_index()
        models = models_json["models"]
        filenames = [model['filename'] for model in models]
        descriptions = [model['details'] for model in models]
    return pd.DataFrame({'model': filenames, 'description': descriptions})

class AnnotationResult():
    """
    Class that represents the result of a celltyping annotation process.
    Parameters
    ----------
    labels
        A :class:`~pandas.DataFrame` object returned from the celltyping process, showing the predicted labels.
    decision_mat
        A :class:`~pandas.DataFrame` object returned from the celltyping process, showing the decision matrix.
    prob_mat
        A :class:`~pandas.DataFrame` object returned from the celltyping process, showing the probability matrix.
    adata
        An :class:`~anndata.AnnData` object representing the input object.
        
    Attributes
    ----------
    predicted_labels
        Predicted labels including the individual prediction results and (if majority voting is done) majority voting results.
    decision_matrix
        Decision matrix with the decision score of each cell belonging to a given cell type.
    probability_matrix
        Probability matrix representing the probability each cell belongs to a given cell type (transformed from decision matrix by the sigmoid function).
    cell_count
        Number of input cells which undergo the prediction process.
    adata
        An :class:`~anndata.AnnData` object representing the input data.
    """
    def __init__(self, labels: pd.DataFrame, decision_mat: pd.DataFrame, prob_mat: pd.DataFrame, adata: AnnData):
        self.predicted_labels = labels
        self.decision_matrix = decision_mat
        self.probability_matrix = prob_mat
        self.adata = adata
        self.cell_count = labels.shape[0]

    def summary_frequency(self, by: str = 'predicted_labels') -> pd.DataFrame:
        """
        Get the frequency of cells belonging to each cell type predicted by celltypist.
        Parameters
        
        ----------
        by: str
            Column name of :attr:`~celltypist.classifier.AnnotationResult.predicted_labels` specifying the prediction type which the summary is based on.
            Set to `'majority_voting'` if you want to summarize for the majority voting classifier.
            (Default: `'predicted_labels'`)
            
        Returns
        ----------
        :class:`~pandas.DataFrame`
            A :class:`~pandas.DataFrame` object with cell type frequencies.
        """
        unique, counts = np.unique(self.predicted_labels[by], return_counts=True)
        df = pd.DataFrame(list(zip(unique, counts)), columns=["celltype", "counts"])
        df.sort_values(['counts'], ascending=False, inplace=True)
        return df

    def to_adata(self, insert_labels: bool = True, insert_conf: bool = True, insert_conf_by: str = 'predicted_labels', insert_decision: bool = False, insert_prob: bool = False, prefix: str = '') -> AnnData:
        """
        Insert the predicted labels, decision or probability matrix, and (if majority voting is done) majority voting results into the AnnData object.
        
        Parameters
        ----------
        insert_labels: bool
            Whether to insert the predicted cell type labels and (if majority voting is done) majority voting-based labels into the AnnData object.
            (Default: `True`)
        insert_conf: bool
            Whether to insert the confidence scores into the AnnData object.
            (Default: `True`)
        insert_conf_by: str
            Column name of :attr:`~celltypist.classifier.AnnotationResult.predicted_labels` specifying the prediction type which the confidence scores are based on.
            Setting to `'majority_voting'` will insert the confidence scores corresponding to the majority-voting result.
            (Default: `'predicted_labels'`)
        insert_decision: bool
            Whether to insert the decision matrix into the AnnData object.
            (Default: `False`)
        insert_prob: bool
            Whether to insert the probability matrix into the AnnData object. This will override the decision matrix even when `insert_decision` is set to `True`.
            (Default: `False`)
        prefix:  str
            Prefix for the inserted columns in the AnnData object. Default to no prefix used.
            
        Returns
        ----------
        :class:`~anndata.AnnData`
            Depending on whether majority voting is done, an :class:`~anndata.AnnData` object with the following columns (prefixed with `prefix`) added to the observation metadata:
            1) **predicted_labels**, individual prediction outcome for each cell.
            2) **over_clustering**, over-clustering result for the cells.
            3) **majority_voting**, the cell type label assigned to each cell after the majority voting process.
            4) **conf_score**, the confidence score of each cell.
            5) **name of each cell type**, which represents the decision scores (or probabilities if `insert_prob` is `True`) of a given cell type across cells.
        """
        if insert_labels:
            self.adata.obs[[f"{prefix}{x}" for x in self.predicted_labels.columns]] = self.predicted_labels
        if insert_conf:
            if insert_conf_by == 'predicted_labels':
                self.adata.obs[f"{prefix}conf_score"] = self.probability_matrix.max(axis=1).values
            elif insert_conf_by == 'majority_voting':
                if insert_conf_by not in self.predicted_labels:
                    raise KeyError(f"?? Did not find the column `majority_voting` in the `AnnotationResult.predicted_labels`, perform majority voting beforehand or use `insert_conf_by = 'predicted_labels'` instead")
                self.adata.obs[f"{prefix}conf_score"] = [row[self.predicted_labels.majority_voting[index]] for index, row in self.probability_matrix.iterrows()]
            else:
                raise KeyError(f"?? Unrecognized `insert_conf_by` value ('{insert_conf_by}'), should be one of `'predicted_labels'` or `'majority_voting'`")
        if insert_prob:
            self.adata.obs[[f"{prefix}{x}" for x in self.probability_matrix.columns]] = self.probability_matrix
        elif insert_decision:
            self.adata.obs[[f"{prefix}{x}" for x in self.decision_matrix.columns]] = self.decision_matrix
        return self.adata

    def to_plots(self, folder: str, plot_probability: bool = False, format: str = 'pdf', prefix: str = '') -> None:
        """
        Plot the celltyping and (if majority voting is done) majority-voting results.
        
        Parameters
        ----------
        folder: str
            Path to a folder which stores the output figures.
        plot_probability:bool
            Whether to also plot the decision score and probability distributions of each cell type across the test cells.
            If `True`, a number of figures will be generated (may take some time if the input data is large).
            (Default: `False`)
        format: str
            Format of output figures. Default to vector PDF files (note dots are still drawn with png backend).
            (Default: `'pdf'`)
        prefix: str
            Prefix for the output figures. Default to no prefix used.
            
        Returns
        ----------
        None
            Depending on whether majority voting is done and `plot_probability`, multiple UMAP plots showing the prediction and majority voting results in the `folder`:
            1) **predicted_labels**, individual prediction outcome for each cell overlaid onto the UMAP.
            2) **over_clustering**, over-clustering result of the cells overlaid onto the UMAP.
            3) **majority_voting**, the cell type label assigned to each cell after the majority voting process overlaid onto the UMAP.
            4) **name of each cell type**, which represents the decision scores and probabilities of a given cell type distributed across cells overlaid onto the UMAP.
        """
        if not os.path.isdir(folder):
            raise FileNotFoundError(f"?? Output folder {folder} does not exist. Please provide a valid folder")
        if 'X_umap' in self.adata.obsm:
            logger.info("?? Detected existing UMAP coordinates, will plot the results accordingly")
        elif 'connectivities' in self.adata.obsp:
            logger.info("?? Generating UMAP coordinates based on the neighborhood graph")
            sc.tl.umap(self.adata)
        else:
            logger.info("?? Constructing the neighborhood graph and generating UMAP coordinates")
            adata = self.adata.copy()
            self.adata.obsm['X_pca'], self.adata.obsp['connectivities'], self.adata.obsp['distances'], self.adata.uns['neighbors'] = Classifier._construct_neighbor_graph(adata)
            sc.tl.umap(self.adata)
        logger.info("?? Plotting the results")
        sc.settings.set_figure_params(figsize=[6.4, 6.4], format=format)
        self.adata.obs[self.predicted_labels.columns] = self.predicted_labels
        for column in self.predicted_labels:
            sc.pl.umap(self.adata, color = column, legend_loc = 'on data', show = False, legend_fontweight = 'normal', title = column.replace('_', ' '))
            plt.savefig(os.path.join(folder, prefix + column + '.' + format))
        if plot_probability:
            for column in self.probability_matrix:
                self.adata.obs['decision score'] = self.decision_matrix[column]
                self.adata.obs['probability'] = self.probability_matrix[column]
                sc.pl.umap(self.adata, color = ['decision score', 'probability'], show = False)
                plt.savefig(os.path.join(folder, prefix + column.replace('/','_') + '.' + format))
            self.adata.obs.drop(columns=['decision score', 'probability'], inplace=True)

    def to_table(self, folder: str, prefix: str = '', xlsx: bool = False) -> None:
        """
        Write out tables of predicted labels, decision matrix, and probability matrix.
        
        Parameters
        ----------
        folder: str
            Path to a folder which stores the output table/tables.
        prefix: str
            Prefix for the output table/tables. Default to no prefix used.
        xlsx: bool
            Whether to merge output tables into a single Excel (.xlsx).
            (Default: `False`)
            
        Returns
        ----------
        None
            Depending on `xlsx`, return table(s) of predicted labels, decision matrix and probability matrix.
        """
        if not os.path.isdir(folder):
            raise FileNotFoundError(f"?? Output folder {folder} does not exist. Please provide a valid folder")
        if not xlsx:
            self.predicted_labels.to_csv(os.path.join(folder, f"{prefix}predicted_labels.csv"))
            self.decision_matrix.to_csv(os.path.join(folder, f"{prefix}decision_matrix.csv"))
            self.probability_matrix.to_csv(os.path.join(folder, f"{prefix}probability_matrix.csv"))
        else:
            with pd.ExcelWriter(os.path.join(folder, f"{prefix}annotation_result.xlsx")) as writer:
                self.predicted_labels.to_excel(writer, sheet_name="Predicted Labels")
                self.decision_matrix.to_excel(writer, sheet_name="Decision Matrix")
                self.probability_matrix.to_excel(writer, sheet_name="Probability Matrix")

    def __repr__(self):
        base = f"CellTypist prediction result for {self.cell_count} query cells"
        base += f"\n    predicted_labels: data frame with {self.predicted_labels.shape[1]} {'columns' if self.predicted_labels.shape[1] > 1 else 'column'} ({str(list(self.predicted_labels.columns))[1:-1]})"
        base += f"\n    decision_matrix: data frame with {self.cell_count} query cells and {self.decision_matrix.shape[1]} cell types"
        base += f"\n    probability_matrix: data frame with {self.cell_count} query cells and {self.probability_matrix.shape[1]} cell types"
        base += f"\n    adata: AnnData object referred"
        return base

class Classifier():
    """
    Class that wraps the celltyping and majority voting processes.
    Parameters
    ----------
    filename: Union[AnnData,str]
        Path to the input count matrix (supported types are csv, txt, tsv, tab and mtx) or AnnData object (h5ad).
        If it's the former, a cell-by-gene format is desirable (see `transpose` for more information).
        Also accepts the input as an :class:`~anndata.AnnData` object already loaded in memory.
        Genes should be gene symbols. Non-expressed genes are preferred to be provided as well.
    model: Union[Model,str]
        A :class:`~celltypist.models.Model` object that wraps the logistic Classifier and the StandardScaler, the
        path to the desired model file, or the model name.
    transpose: bool
        Whether to transpose the input matrix. Set to `True` if `filename` is provided in a gene-by-cell format.
        (Default: `False`)
    gene_file: Optional[str]
        Path to the file which stores each gene per line corresponding to the genes used in the provided mtx file.
        Ignored if `filename` is not provided in the mtx format.
    cell_file: Optional[str]
        Path to the file which stores each cell per line corresponding to the cells used in the provided mtx file.
        Ignored if `filename` is not provided in the mtx format.
    Attributes
    ----------
    filename
        Path to the input dataset. This attribute exists only when the input is a file path.
    adata
        An :class:`~anndata.AnnData` object which stores the log1p normalized expression data in `.X` or `.raw.X`.
    indata
        The expression matrix used for predictions stored in the log1p normalized format.
    indata_genes
        All the genes included in the input data.
    indata_names
        All the cells included in the input data.
    model
        A :class:`~celltypist.models.Model` object that wraps the logistic Classifier and the StandardScaler.
    """
    def __init__(self, filename: Union[AnnData,str] = "", model: Union[Model,str] = "", transpose: bool = False, gene_file: Optional[str] = None, cell_file: Optional[str] = None, check_expression: bool = False, ):
        if isinstance(model, str):
            model = Model.load(model)
        self.model = model
        if not filename:
            logger.warn(f"?? No input file provided to the classifier")
            return
        if isinstance(filename, str):
            self.filename = filename
            logger.info(f"?? Input file is '{self.filename}'")
            logger.info(f"? Loading data")
        if isinstance(filename, str) and filename.endswith(('.csv', '.txt', '.tsv', '.tab', '.mtx', '.mtx.gz')):
            self.adata = sc.read(self.filename)
            if transpose:
                self.adata = self.adata.transpose()
            if self.filename.endswith(('.mtx', '.mtx.gz')):
                if (gene_file is None) or (cell_file is None):
                    raise FileNotFoundError("?? Missing `gene_file` and/or `cell_file`. Please provide both arguments together with the input mtx file")
                genes_mtx = pd.read_csv(gene_file, header=None)[0].values
                cells_mtx = pd.read_csv(cell_file, header=None)[0].values
                if len(genes_mtx) != self.adata.n_vars:
                    raise ValueError(f"?? The number of genes in {gene_file} does not match the number of genes in {self.filename}")
                if len(cells_mtx) != self.adata.n_obs:
                    raise ValueError(f"?? The number of cells in {cell_file} does not match the number of cells in {self.filename}")
                self.adata.var_names = genes_mtx
                self.adata.obs_names = cells_mtx
            self.adata.var_names_make_unique()
            if not float(self.adata.X.max()).is_integer():
                logger.warn(f"?? Warning: the input file seems not a raw count matrix. The prediction result may be biased")
            if (self.adata.n_vars >= 80000) or (len(self.adata.var_names[0]) >= 30) or (len(self.adata.obs_names.intersection(pd.Index(['GAPDH', 'ACTB', 'CALM1', 'PTPRC']))) >= 1):
                raise ValueError(f"?? The input matrix is detected to be a gene-by-cell matrix. Please provide a cell-by-gene matrix or add the input transpose option")
            sc.pp.normalize_total(self.adata, target_sum=1e4)
            sc.pp.log1p(self.adata)
            self.indata = self.adata.X
            self.indata_genes = self.adata.var_names
            self.indata_names = self.adata.obs_names
        elif isinstance(filename, AnnData) or (isinstance(filename, str) and filename.endswith('.h5ad')):
            self.adata = sc.read(filename) if isinstance(filename, str) else filename
            self.adata.var_names_make_unique()
            if self.adata.X.min() < 0:
                logger.info("?? Detected scaled expression in the input data, will try the `.raw` attribute")
                try:
                    self.indata = self.adata.raw.X
                    self.indata_genes = self.adata.raw.var_names
                    self.indata_names = self.adata.raw.obs_names
                except Exception as e:
                    raise Exception(f"?? Fail to use the `.raw` attribute in the input object. {e}")
            else:
                self.indata = self.adata.X
                self.indata_genes = self.adata.var_names
                self.indata_names = self.adata.obs_names
            if check_expression and np.abs(np.expm1(self.indata[0]).sum()-10000) > 1:
                raise ValueError("?? Invalid expression matrix, expect log1p normalized expression to 10000 counts per cell")
        else:
            raise ValueError("?? Invalid input. Supported types: .csv, .txt, .tsv, .tab, .mtx, .mtx.gz and .h5ad, or AnnData loaded in memory")

        logger.info(f"?? Input data has {self.indata.shape[0]} cells and {len(self.indata_genes)} genes")

    def celltype(self, mode: str = 'best match', p_thres: float = 0.5) -> AnnotationResult:
        """
        Run celltyping jobs to predict cell types of input data.
        
        Parameters
        ----------
        mode: str
            The way cell prediction is performed.
            For each query cell, the default (`'best match'`) is to choose the cell type with the largest score/probability as the final prediction.
            Setting to `'prob match'` will enable a multi-label classification, which assigns 0 (i.e., unassigned), 1, or >=2 cell type labels to each query cell.
            (Default: `'best match'`)
        p_thres: str
            Probability threshold for the multi-label classification. Ignored if `mode` is `'best match'`.
            (Default: 0.5)
            
        Returns
        ----------
        :class:`~celltypist.classifier.AnnotationResult`
            An :class:`~celltypist.classifier.AnnotationResult` object. Four important attributes within this class are:
            1) :attr:`~celltypist.classifier.AnnotationResult.predicted_labels`, predicted labels from celltypist.
            2) :attr:`~celltypist.classifier.AnnotationResult.decision_matrix`, decision matrix from celltypist.
            3) :attr:`~celltypist.classifier.AnnotationResult.probability_matrix`, probability matrix from celltypist.
            4) :attr:`~celltypist.classifier.AnnotationResult.adata`, AnnData object representation of the input data.
        """
        logger.info(f"?? Matching reference genes in the model")
        k_x = np.isin(self.indata_genes, self.model.classifier.features)
        if k_x.sum() == 0:
            raise ValueError(f"?? No features overlap with the model. Please provide gene symbols")
        else:
            logger.info(f"?? {k_x.sum()} features used for prediction")
        k_x_idx = np.where(k_x)[0]
        #self.indata = self.indata[:, k_x_idx]
        self.indata_genes = self.indata_genes[k_x_idx]
        lr_idx = pd.DataFrame(self.model.classifier.features, columns=['features']).reset_index().set_index('features').loc[self.indata_genes, 'index'].values

        logger.info(f"?? Scaling input data")
        means_ = self.model.scaler.mean_[lr_idx]
        sds_ = self.model.scaler.scale_[lr_idx]
        self.indata = (self.indata[:, k_x_idx] - means_) / sds_
        self.indata[self.indata > 10] = 10

        ni, fs, cf = self.model.classifier.n_features_in_, self.model.classifier.features, self.model.classifier.coef_
        self.model.classifier.n_features_in_ = lr_idx.size
        self.model.classifier.features = self.model.classifier.features[lr_idx]
        self.model.classifier.coef_ = self.model.classifier.coef_[:, lr_idx]

        logger.info("??? Predicting labels")
        decision_mat, prob_mat, lab = self.model.predict_labels_and_prob(self.indata, mode = mode, p_thres = p_thres)
        logger.info("? Prediction done!")

        #restore model after prediction
        self.model.classifier.n_features_in_, self.model.classifier.features, self.model.classifier.coef_ = ni, fs, cf

        cells = self.indata_names
        return AnnotationResult(pd.DataFrame(lab, columns=['predicted_labels'], index=cells, dtype='category'), pd.DataFrame(decision_mat, columns=self.model.classifier.classes_, index=cells), pd.DataFrame(prob_mat, columns=self.model.classifier.classes_, index=cells), self.adata)

    @staticmethod
    def _construct_neighbor_graph(adata: AnnData) -> tuple:
        """Construct a neighborhood graph. This function is for internal use."""
        if 'X_pca' not in adata.obsm.keys():
            if adata.X.min() < 0:
                adata = adata.raw.to_adata()
            if 'highly_variable' not in adata.var:
                sc.pp.filter_genes(adata, min_cells=5)
                sc.pp.highly_variable_genes(adata, n_top_genes = min([2500, adata.n_vars]))
            adata = adata[:, adata.var.highly_variable]
            sc.pp.scale(adata, max_value=10)
            sc.tl.pca(adata, n_comps=50)
        sc.pp.neighbors(adata, n_neighbors=10, n_pcs=50)
        return adata.obsm['X_pca'], adata.obsp['connectivities'], adata.obsp['distances'], adata.uns['neighbors']

    def over_cluster(self, resolution: Optional[float] = None) -> pd.Series:
        """
        Over-clustering input data with a canonical Scanpy pipeline. A neighborhood graph will be used (or constructed if not found) for the over-clustering.
        
        Parameters
        ----------
        resolution: Optional[float]
            Resolution parameter for leiden clustering which controls the coarseness of the clustering.
            Default to 5, 10, 15, 20, 25 and 30 for datasets with cell numbers less than 5k, 20k, 40k, 100k, 200k and above, respectively.
            
        Returns
        ----------
        :class:`~pandas.Series`
            A :class:`~pandas.Series` object showing the over-clustering result.
        """
        if 'connectivities' not in self.adata.obsp:
            logger.info("?? Can not detect a neighborhood graph, will construct one before the over-clustering")
            adata = self.adata.copy()
            self.adata.obsm['X_pca'], self.adata.obsp['connectivities'], self.adata.obsp['distances'], self.adata.uns['neighbors'] = Classifier._construct_neighbor_graph(adata)
        else:
            logger.info("?? Detected a neighborhood graph in the input object, will run over-clustering on the basis of it")
        if resolution is None:
            if self.adata.n_obs < 5000:
                resolution = 5
            elif self.adata.n_obs < 20000:
                resolution = 10
            elif self.adata.n_obs < 40000:
                resolution = 15
            elif self.adata.n_obs < 100000:
                resolution = 20
            elif self.adata.n_obs < 200000:
                resolution = 25
            else:
                resolution = 30
        logger.info(f"?? Over-clustering input data with resolution set to {resolution}")
        sc.tl.leiden(self.adata, resolution=resolution, key_added='over_clustering')
        return self.adata.obs.pop('over_clustering')

    @staticmethod
    def majority_vote(predictions: AnnotationResult, over_clustering: Union[list, tuple, np.ndarray, pd.Series, pd.Index], min_prop: float = 0) -> AnnotationResult:
        """
        Majority vote the celltypist predictions using the result from the over-clustering.
        
        Parameters
        ----------
        predictions: AnnotationResult
            An :class:`~celltypist.classifier.AnnotationResult` object containing the :attr:`~celltypist.classifier.AnnotationResult.predicted_labels`.
        over_clustering: Union[list, tuple, np.ndarray, pd.Series, pd.Index]
            A list, tuple, numpy array, pandas series or index containing the over-clustering information.
        min_prop: float
            For the dominant cell type within a subcluster, the minimum proportion of cells required to support naming of the subcluster by this cell type.
            (Default: 0)
            
        Returns
        ----------
        output:class:`~celltypist.classifier.AnnotationResult`
            An :class:`~celltypist.classifier.AnnotationResult` object. Four important attributes within this class are:
            1) :attr:`~celltypist.classifier.AnnotationResult.predicted_labels`, predicted labels from celltypist.
            2) :attr:`~celltypist.classifier.AnnotationResult.decision_matrix`, decision matrix from celltypist.
            3) :attr:`~celltypist.classifier.AnnotationResult.probability_matrix`, probability matrix from celltypist.
            4) :attr:`~celltypist.classifier.AnnotationResult.adata`, AnnData object representation of the input data.
        """
        if isinstance(over_clustering, (list, tuple)):
            over_clustering = np.array(over_clustering)
        logger.info("??? Majority voting the predictions")
        votes = pd.crosstab(predictions.predicted_labels['predicted_labels'], over_clustering)
        majority = votes.idxmax(axis=0)
        freqs = (votes / votes.sum(axis=0).values).max(axis=0)
        majority[freqs < min_prop] = 'Heterogeneous'
        majority = majority[over_clustering].reset_index()
        majority.index = predictions.predicted_labels.index
        majority.columns = ['over_clustering', 'majority_voting']
        majority['majority_voting'] = majority['majority_voting'].astype('category')
        predictions.predicted_labels = predictions.predicted_labels.join(majority)
        logger.info("? Majority voting done!")
        return predictions


class Celltypist():
    """
    Build the ACTINN model
    
    ----------
    Parameters
    classifier: Classification function
        Class that wraps the celltyping and majority voting processes, as defined above
    scaler: StandardScaler
        The scale factor for normalization.
    description: str
        text description of the model, otherwise it will be set as {'date': date, 'details': details, 'url': url, 'source': source, 'version': version, 'number_celltypes': len(classifier.classes_)} in fit function
        
    """
    def __init__(self, clf=None, scaler=None, description=None):
        self.classifier = clf
        self.scaler = scaler
        self.description = description
        
    def fit(self, X = None,
              labels: Optional[Union[str, list, tuple, np.ndarray, pd.Series, pd.Index]] = None,
              genes: Optional[Union[str, list, tuple, np.ndarray, pd.Series, pd.Index]] = None,
              transpose_input: bool = False,
              check_expression: bool = True,
              #LR param
              C: float = 1.0, solver: Optional[str] = None, max_iter: int = 1000, n_jobs: Optional[int] = None,
              #SGD param
              use_SGD: bool = False, alpha: float = 0.0001,
              #mini-batch
              mini_batch: bool = False, batch_number: int = 100, batch_size: int = 1000, epochs: int = 10, balance_cell_type: bool = False,
              #feature selection
              feature_selection: bool = False, top_genes: int = 300,
              #description
              date: str = '', details: str = '', url: str = '', source: str = '', version: str = '',
              #other param
              **kwargs
             ):
        """  
        Train a celltypist model using mini-batch (optional) logistic classifier with a global solver or stochastic gradient descent (SGD) learning.
        
        Parameters
        ----------
        X: Path
            Path to the input count matrix (supported types are csv, txt, tsv, tab and mtx) or AnnData (h5ad).
            Also accepts the input as an :class:`~anndata.AnnData` object, or any array-like objects already loaded in memory.
            See `check_expression` for detailed format requirements.
            A cell-by-gene format is desirable (see `transpose_input` for more information).
        labels: Optional[Union[str, list, tuple, np.ndarray, pd.Series, pd.Index]]
            Path to the file containing cell type label per line corresponding to the cells in `X`.
            Also accepts any list-like objects already loaded in memory (such as an array).
            If `X` is specified as an AnnData, this argument can also be set as a column name from cell metadata.
        genes: Optional[Union[str, list, tuple, np.ndarray, pd.Series, pd.Index]]
            Path to the file containing one gene per line corresponding to the genes in `X`.
            Also accepts any list-like objects already loaded in memory (such as an array).
            Note `genes` will be extracted from `X` where possible (e.g., `X` is an AnnData or data frame).
        transpose_input: bool
            Whether to transpose the input matrix. Set to `True` if `X` is provided in a gene-by-cell format.
            (Default: `False`)
        check_expression: bool
            Check whether the expression matrix in the input data is supplied as required.
            Except the case where a path to the raw count table file is specified, all other inputs for `X` should be in log1p normalized expression to 10000 counts per cell.
            Set to `False` if you want to train the data regardless of the expression formats.
            (Default: `True`)
        C: float
            Inverse of L2 regularization strength for traditional logistic classifier. A smaller value can possibly improve model generalization while at the cost of decreased accuracy.
            This argument is ignored if SGD learning is enabled (`use_SGD = True`).
            (Default: 1.0)
        solver: Optional[str]
            Algorithm to use in the optimization problem for traditional logistic classifier.
            The default behavior is to choose the solver according to the size of the input data.
            This argument is ignored if SGD learning is enabled (`use_SGD = True`).
        max_iter: int
            Maximum number of iterations before reaching the minimum of the cost function.
            Try to decrease `max_iter` if the cost function does not converge for a long time.
            This argument is for both traditional and SGD logistic classifiers, and will be ignored if mini-batch SGD training is conducted (`use_SGD = True` and `mini_batch = True`).
            (Default: 1000)
        n_jobs: Optional[int]
            Number of CPUs used. Default to one CPU. `-1` means all CPUs are used.
            This argument is for both traditional and SGD logistic classifiers.
        use_SGD: bool
            Whether to implement SGD learning for the logistic classifier.
            (Default: `False`)
        alpha: float
            L2 regularization strength for SGD logistic classifier. A larger value can possibly improve model generalization while at the cost of decreased accuracy.
            This argument is ignored if SGD learning is disabled (`use_SGD = False`).
            (Default: 0.0001)
        mini_batch: bool
            Whether to implement mini-batch training for the SGD logistic classifier.
            Setting to `True` may improve the training efficiency for large datasets (for example, >100k cells).
            This argument is ignored if SGD learning is disabled (`use_SGD = False`).
            (Default: `False`)
        batch_number: int
            The number of batches used for training in each epoch. Each batch contains `batch_size` cells.
            For datasets which cannot be binned into `batch_number` batches, all batches will be used.
            This argument is relevant only if mini-batch SGD training is conducted (`use_SGD = True` and `mini_batch = True`).
            (Default: 100)
        batch_size: int
            The number of cells within each batch.
            This argument is relevant only if mini-batch SGD training is conducted (`use_SGD = True` and `mini_batch = True`).
            (Default: 1000)
        epochs: int
            The number of epochs for the mini-batch training procedure.
            The default values of `batch_number`, `batch_size`, and `epochs` together allow observing ~10^6 training cells.
            This argument is relevant only if mini-batch SGD training is conducted (`use_SGD = True` and `mini_batch = True`).
            (Default: 10)
        balance_cell_type: bool
            Whether to balance the cell type frequencies in mini-batches during each epoch.
            Setting to `True` will sample rare cell types with a higher probability, ensuring close-to-even cell type distributions in mini-batches.
            This argument is relevant only if mini-batch SGD training is conducted (`use_SGD = True` and `mini_batch = True`).
            (Default: `False`)
        feature_selection: bool
            Whether to perform two-pass data training where the first round is used for selecting important features/genes using SGD learning.
            If `True`, the training time will be longer.
            (Default: `False`)
        top_genes: int
            The number of top genes selected from each class/cell-type based on their absolute regression coefficients.
            The final feature set is combined across all classes (i.e., union).
            (Default: 300)
        date: str
            Free text of the date of the model. Default to the time when the training is completed.
        details: str
            Free text of the description of the model.
        url: str
            Free text of the (possible) download url of the model.
        source: str
            Free text of the source (publication, database, etc.) of the model.
        version: str
            Free text of the version of the model.
        **kwargs
            Other keyword arguments passed to :class:`~sklearn.linear_model.LogisticRegression` (`use_SGD = False`) or :class:`~sklearn.linear_model.SGDClassifier` (`use_SGD = True`).
            
        Returns
        ----------
        :class:`~celltypist.models.Model`
            An instance of the :class:`~celltypist.models.Model` trained by celltypist.
        """
        #prepare
        logger.info("?? Preparing data before training")
        indata, labels, genes = prepare_data_celltypist(X, labels, genes, transpose_input)
        indata = to_array_celltypist(indata)
        labels = np.array(labels)
        genes = np.array(genes)
        #check
        if check_expression and (np.abs(np.expm1(indata[0]).sum()-10000) > 1):
            raise ValueError("?? Invalid expression matrix, expect log1p normalized expression to 10000 counts per cell")
        if len(labels) != indata.shape[0]:
            raise ValueError(f"?? Length of training labels ({len(labels)}) does not match the number of input cells ({indata.shape[0]})")
        if len(genes) != indata.shape[1]:
            raise ValueError(f"?? The number of genes ({len(genes)}) provided does not match the number of genes in the training data ({indata.shape[1]})")
        #filter
        flag = indata.sum(axis = 0) == 0
        if flag.sum() > 0:
            logger.info(f"?? {flag.sum()} non-expressed genes are filtered out")
            #indata = indata[:, ~flag]
            genes = genes[~flag]
        #scaler
        logger.info(f"?? Scaling input data")
        scaler = StandardScaler()
        indata = scaler.fit_transform(indata[:, ~flag] if flag.sum() > 0 else indata)
        indata[indata > 10] = 10
        #classifier
        if use_SGD or feature_selection:
            classifier = SGDClassifier_celltypist(indata = indata, labels = labels, alpha = alpha, max_iter = max_iter, n_jobs = n_jobs, mini_batch = mini_batch, batch_number = batch_number, batch_size = batch_size, epochs = epochs, balance_cell_type = balance_cell_type, **kwargs)
        else:
            classifier = LRClassifier_celltypist(indata = indata, labels = labels, C = C, solver = solver, max_iter = max_iter, n_jobs = n_jobs, **kwargs)
        #feature selection -> new classifier and scaler
        if feature_selection:
            logger.info(f"?? Selecting features")
            if len(genes) <= top_genes:
                raise ValueError(f"?? The number of genes ({len(genes)}) is fewer than the `top_genes` ({top_genes}). Unable to perform feature selection")
            gene_index = np.argpartition(np.abs(classifier.coef_), -top_genes, axis = 1)[:, -top_genes:]
            gene_index = np.unique(gene_index)
            logger.info(f"?? {len(gene_index)} features are selected")
            genes = genes[gene_index]
            #indata = indata[:, gene_index]
            logger.info(f"??? Starting the second round of training")
            if use_SGD:
                classifier = SGDClassifier_celltypist(indata = indata[:, gene_index], labels = labels, alpha = alpha, max_iter = max_iter, n_jobs = n_jobs, mini_batch = mini_batch, batch_number = batch_number, batch_size = batch_size, epochs = epochs, balance_cell_type = balance_cell_type, **kwargs)
            else:
                classifier = LRClassifier_celltypist(indata = indata[:, gene_index], labels = labels, C = C, solver = solver, max_iter = max_iter, n_jobs = n_jobs, **kwargs)
            scaler.mean_ = scaler.mean_[gene_index]
            scaler.var_ = scaler.var_[gene_index]
            scaler.scale_ = scaler.scale_[gene_index]
            scaler.n_features_in_ = len(gene_index)
        #model finalization
        classifier.features = genes
        if not date:
            date = str(datetime.now())
        description = {'date': date, 'details': details, 'url': url, 'source': source, 'version': version, 'number_celltypes': len(classifier.classes_)}
        logger.info(f"? Model training done!")
        
        self.classifier = classifier
        self.scaler = scaler
        self.description = description
        
    def predict(self, filename: Union[AnnData,str] = "",
                 check_expression: bool = False,
                 load_model: bool = False,
                 model: Optional[Union[str, Model]] = None,
                 transpose_input: bool = False,
                 gene_file: Optional[str] = None,
                 cell_file: Optional[str] = None,
                 mode: str = 'best match',
                 p_thres: float = 0.5,
                 majority_voting: bool = False,
                 over_clustering: Optional[Union[str, list, tuple, np.ndarray, pd.Series, pd.Index]] = None,
                 min_prop: float = 0) -> AnnotationResult:
        """
        Run the prediction and (optional) majority voting to annotate the input dataset.
        
        Parameters
        ----------
        filename: Union[AnnData,str]
            Path to the input count matrix (supported types are csv, txt, tsv, tab and mtx) or AnnData (h5ad).
            If it's the former, a cell-by-gene format is desirable (see `transpose_input` for more information).
            Also accepts the input as an :class:`~anndata.AnnData` object already loaded in memory.
            Genes should be gene symbols. Non-expressed genes are preferred to be provided as well.
        model: Optional[Union[str, Model]]
            Model used to predict the input cells. Default to using the `'Immune_All_Low.pkl'` model.
            Can be a :class:`~celltypist.models.Model` object that wraps the logistic Classifier and the StandardScaler, the
            path to the desired model file, or the model name.
            To see all available models and their descriptions, use :func:`~celltypist.models.models_description`.
        transpose_input: bool
            Whether to transpose the input matrix. Set to `True` if `filename` is provided in a gene-by-cell format.
            (Default: `False`)
        gene_file: Optional[str]
            Path to the file which stores each gene per line corresponding to the genes used in the provided mtx file.
            Ignored if `filename` is not provided in the mtx format.
        cell_file: Optional[str]
            Path to the file which stores each cell per line corresponding to the cells used in the provided mtx file.
            Ignored if `filename` is not provided in the mtx format.
        mode: str
            The way cell prediction is performed.
            For each query cell, the default (`'best match'`) is to choose the cell type with the largest score/probability as the final prediction.
            Setting to `'prob match'` will enable a multi-label classification, which assigns 0 (i.e., unassigned), 1, or >=2 cell type labels to each query cell.
            (Default: `'best match'`)
        p_thres: float
            Probability threshold for the multi-label classification. Ignored if `mode` is `'best match'`.
            (Default: 0.5)
        majority_voting: bool
            Whether to refine the predicted labels by running the majority voting classifier after over-clustering.
            (Default: `False`)
        over_clustering: Optional[Union[str, list, tuple, np.ndarray, pd.Series, pd.Index]]
            This argument can be provided in several ways:
            1) an input plain file with the over-clustering result of one cell per line.
            2) a string key specifying an existing metadata column in the AnnData (pre-created by the user).
            3) a python list, tuple, numpy array, pandas series or index representing the over-clustering result of the input cells.
            4) if none of the above is provided, will use a heuristic over-clustering approach according to the size of input data.
            Ignored if `majority_voting` is set to `False`.
        min_prop: float
            For the dominant cell type within a subcluster, the minimum proportion of cells required to support naming of the subcluster by this cell type.
            Ignored if `majority_voting` is set to `False`.
            Subcluster that fails to pass this proportion threshold will be assigned `'Heterogeneous'`.
            (Default: 0)
            
        Returns
        ----------
        output :class:`~celltypist.classifier.AnnotationResult`
            An :class:`~celltypist.classifier.AnnotationResult` object. Four important attributes within this class are:
            1) :attr:`~celltypist.classifier.AnnotationResult.predicted_labels`, predicted labels from celltypist.
            2) :attr:`~celltypist.classifier.AnnotationResult.decision_matrix`, decision matrix from celltypist.
            3) :attr:`~celltypist.classifier.AnnotationResult.probability_matrix`, probability matrix from celltypist.
            4) :attr:`~celltypist.classifier.AnnotationResult.adata`, AnnData representation of the input data.
        """
        #load model
        # lr_classifier = Model(self.classifier, self.scaler, self.description) if isinstance(model, Model) else Model.load(model)
        if load_model:
            lr_classifier = Model.load(model)
        else: 
            lr_classifier = Model(self.classifier, self.scaler, self.description)
        #construct Classifier class
        clf = Classifier(filename = filename, model = lr_classifier, transpose = transpose_input, gene_file = gene_file, cell_file = cell_file,check_expression = check_expression)
        #predict
        predictions = clf.celltype(mode = mode, p_thres = p_thres)
        if not majority_voting:
            return predictions
        if predictions.cell_count <= 50:
            logger.warn(f"?? Warning: the input number of cells ({predictions.cell_count}) is too few to conduct proper over-clustering; no majority voting is performed")
            return predictions
        #over clustering
        if over_clustering is None:
            over_clustering = clf.over_cluster()
            predictions.adata = clf.adata
        elif isinstance(over_clustering, str):
            if over_clustering in clf.adata.obs:
                over_clustering = clf.adata.obs[over_clustering]
            else:
                logger.info(f"?? Did not identify '{over_clustering}' as a cell metadata column, assume it to be a plain text file")
                try:
                    with open(over_clustering, 'rt') as f:
                        over_clustering = [x.strip() for x in f.readlines()]
                except Exception as e:
                    raise Exception(f"?? {e}")
        if len(over_clustering) != clf.adata.n_obs:
            raise ValueError(f"?? Length of `over_clustering` ({len(over_clustering)}) does not match the number of input cells ({clf.adata.n_obs})")
        #majority voting
        print(predictions)
        return Classifier.majority_vote(predictions, over_clustering, min_prop = min_prop)
    
    def score(self, input_adata, predictions,labels, map, label_conversion=False):
        """
        Run the prediction and (optional) majority voting to evaluate the model performance.
        
        Parameters
        ----------
        input_adata: Anndata
            Input data anndata with label ground truth
        predictions: classifier.AnnotationResult
            Output from prediction function. 
        labels : string
            column name for annotated cell types in the input Anndata
        label_conversion: boolean
            whether to match predicted labels to annotated cell type labels provided in input_adata
        map: dictionary
            a dictionary for label conversion

        Returns
        ----------
        correct: int
            Number of correct predictions
        Accuracy: float
            Prediction accuracy from the model
        """
        pred_labels = np.array(predictions.predicted_labels)[:, 0]
        if label_conversion:
            correct = 0
            #for i, cell in enumerate(np.array(adVal.obs.Cell_type)):
            for i, cell in enumerate(np.array(input_adata.obs[labels])):
                if pred_labels[i] in map[cell]:
                    correct += 1
        else:
            correct = sum(np.array(input_adata.obs[label])==pred_labels)
        return (correct/len(predictions.predicted_labels))


import warnings
warnings.filterwarnings("ignore")
import os
import random
import numpy as np
import pandas as pd
import argparse
import torch
import scanpy as sc

from dance.transforms.preprocess import normalize_adata, filter_data
from dance.transforms.graph_construct import get_adj
from dance.modules.single_modality.clustering.sctag import SCTAG
from dance.datasets.singlemodality import ClusteringDataset

# for repeatability
random.seed(42)
np.random.seed(42)
torch.manual_seed(42)
torch.cuda.manual_seed(42)

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="train", formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--data_dir', default='./data', type=str)
    parser.add_argument('--data_file', default='mouse_bladder_cell',
                        type=str)  # choice=['10X_PBMC', 'mouse_bladder_cell', 'mouse_ES_cell', 'worm_neuron_cell']
    parser.add_argument('--pretrain_file', type=str, default="./sctag_mouse_bladder_cell_pre.pkl")
    parser.add_argument("--k_neighbor", default=15, type=int)
    parser.add_argument("--highly_genes", default=3000, type=int)
    parser.add_argument("--pca_dim", default=50, type=int)
    parser.add_argument("--k", default=2, type=int)
    parser.add_argument("--hidden_dim", default=128, type=int)
    parser.add_argument("--latent_dim", default=15, type=int)
    parser.add_argument("--dec_dim", default=None, type=int)
    parser.add_argument("--dropout", default=0.4, type=float)
    parser.add_argument("--alpha", default=1.0, type=float)
    parser.add_argument("--pretrain_epochs", default=200, type=int)
    parser.add_argument("--epochs", default=500, type=int)
    parser.add_argument("--device", default="cuda")
    parser.add_argument("--W_a", default=0.01, type=float)
    parser.add_argument("--W_x", default=3, type=float)
    parser.add_argument("--W_d", default=0, type=float)
    parser.add_argument("--W_c", default=0.1, type=float)
    parser.add_argument("--lr", default=1e-3, type=float)
    parser.add_argument("--min_dist", default=0.5, type=float)
    parser.add_argument("--max_dist", default=20.0, type=float)
    parser.add_argument("--info_step", default=1, type=int)
    args = parser.parse_args()

    args.pretrain_file = "./sctag_" + args.data_file + "_pre.pkl"
    # Load data
    data = ClusteringDataset(args.data_dir, args.data_file).load_data()
    x = data.X
    y = data.Y

    genes_idx, cells_idx = filter_data(x, highly_genes=args.highly_genes)
    x = x[cells_idx][:, genes_idx]
    y = y[cells_idx]
    n_clusters = len(np.unique(y))

    adata = sc.AnnData(x, dtype=np.float32)
    adata.obs['Group'] = y
    adata = adata.copy()
    adata.obs['DCA_split'] = 'train'
    adata.obs['DCA_split'] = adata.obs['DCA_split'].astype('category')

    adata = normalize_adata(adata,
                            size_factors=True,
                            normalize_input=True,
                            logtrans_input=True)
    x = adata.X
    x_raw = adata.raw.X
    y = adata.obs['Group']
    scale_factor = adata.obs.size_factors

    adj, adj_n = get_adj(x, k=args.k_neighbor, pca_dim=args.pca_dim)

    # Build model & training
    model = SCTAG(x, adj=adj, adj_n=adj_n, n_clusters=n_clusters, k=args.k,
                  hidden_dim=args.hidden_dim, latent_dim=args.latent_dim, dec_dim=args.dec_dim,
                  adj_dim=adj_n.shape[0], dropout=args.dropout, device=args.device, alpha=args.alpha)

    if not os.path.exists(args.pretrain_file):
        model.pre_train(x, x_raw, args.pretrain_file, scale_factor, epochs=args.pretrain_epochs,
                        info_step=args.info_step,
                        W_a=args.W_a, W_x=args.W_x, W_d=args.W_d, min_dist=args.min_dist, max_dist=args.max_dist)
    else:
        print("==> loading checkpoint '{}'".format(args.pretrain_file))
        checkpoint = torch.load(args.pretrain_file)
        model.load_state_dict(checkpoint)

    model.fit(x, x_raw, y, scale_factor, epochs=args.epochs, lr=args.lr,
              W_a=args.W_a, W_x=args.W_x, W_c=args.W_c, info_step=args.info_step)

    y_pred = model.predict()
    #    print(f'Prediction: {y_pred}')

    acc, nmi, ari = model.score(y)
    print("ACC: %.4f, NMI: %.4f, ARI: %.4f" % (acc, nmi, ari))
    print(args)

"""Reproduction information
10X PBMC:
python sctag.py --data_file='10X_PBMC' --W_a=0.01 --W_x=3 --W_c=0.1

Worm Neuron:
python sctag.py --data_file='worm_neuron_cell' --W_a=0.01 --W_x=2 --W_c=0.25

Mouse Bladder:
python sctag.py --data_file='mouse_bladder_cell' --W_a=0.1 --W_x=2.5 --W_c=3

Mouse ES:
python sctag.py --data_file='mouse_ES_cell' --W_a=0.01 --W_x=0.75 --W_c=1
"""


"""
This is the implementation of scTAG.
Extended from https://github.com/Philyzh8/scTAG

Reference
----------
Yu, Z., Y. Lu, Y. Wang, F. Tang, K.-C. Wong, and X. Li. “ZINB-Based Graph Embedding Autoencoder for Single-Cell RNA-Seq Interpretations”. Proceedings of the AAAI Conference on Artificial Intelligence, vol. 36, no. 4, June 2022, pp. 4671-9, doi:10.1609/aaai.v36i4.20392.
"""

from tqdm import tqdm
import numpy as np
import dgl
from dgl.nn import TAGConv
import torch
import torch.nn as nn
from torch.nn import Parameter
import torch.nn.functional as F
import torch.optim as optim
from sklearn import metrics
from sklearn.cluster import KMeans

from dance.utils.loss import ZINBLoss, dist_loss
from dance.utils.metrics import cluster_acc


class SCTAG(nn.Module):
    """scTAG class.

    Parameters
    ----------
    X :
        input features.
    adj :
        adjacency matrix.
    adj_n :
        normalized adjacency matrix.
    n_clusters : int
        number of clusters.
    k : int optional
        number of hops of TAG convolutional layer.
    hidden_dim : int optional
        dimension of hidden layer.
    latent_dim : int optional
        dimension of latent embedding.
    dec_dim : list optional
        dimensions of decoder layers.
    adj_dim : int optional
        dimension of adjacency matrix.
    dropout : float optional
        dropout rate.
    device : str optional
        computing device.
    alpha : float optional
        parameter of soft assign.

    """

    def __init__(self, X, adj, adj_n, n_clusters, k=3, hidden_dim=128, latent_dim=15,
                 dec_dim=None, adj_dim=32, dropout=0.2, device="cuda", alpha=1.0):
        super(SCTAG, self).__init__()
        if dec_dim is None:
            dec_dim = [128, 256, 512]
        self.latent_dim = latent_dim
        self.hidden_dim = hidden_dim
        self.adj = adj
        self.adj_n = adj_n
        self.n_sample = X.shape[0]
        self.in_dim = X.shape[1]
        self.device = device
        self.dropout = dropout
        self.n_clusters = n_clusters
        self.alpha = alpha
        self.k = k
        self.adj_dim = adj_dim
        self.mu = Parameter(torch.Tensor(self.n_clusters, self.latent_dim).to(self.device))

        src, dist = np.nonzero(adj)
        self.G = dgl.graph((src, dist)).to(device)

        src_n, dist_n = np.nonzero(adj_n)
        self.G_n = dgl.graph((src_n, dist_n)).to(device)

        self.encoder1 = TAGConv(self.in_dim, self.hidden_dim, k=k)
        self.encoder2 = TAGConv(self.hidden_dim, self.latent_dim, k=k)
        self.decoderA = DecoderA(latent_dim=self.latent_dim, adj_dim=self.adj_dim,
                                 activation=torch.sigmoid, dropout=self.dropout)
        self.decoderX = DecoderX(self.in_dim, self.latent_dim, n_dec_1=dec_dim[0], n_dec_2=dec_dim[1],
                                 n_dec_3=dec_dim[2])
        self.zinb_loss = ZINBLoss().to(self.device)
        self.to(self.device)

    def forward(self, A_in, X_input):
        """Forward propagation.

        Parameters
        ----------
        A_in :
            input adjacency matrix.
        X_input :
            input features.

        Returns
        -------
        A_out :
            reconstructed adjacency matrix.
        z :
            embedding.
        q :
            soft label.
        _mean :
            data mean from ZINB.
        _disp :
            data dispersion from ZINB.
        _pi :
            data dropout probability from ZINB.

        """
        enc_h = self.encoder1(A_in, X_input)
        z = self.encoder2(A_in, enc_h)
        A_out = self.decoderA(z)
        _mean, _disp, _pi = self.decoderX(z)
        q = self.soft_assign(z)

        return A_out, z, q, _mean, _disp, _pi

    def pre_train(self, x, x_raw, fname, scale_factor, epochs=1000, info_step=10, lr=5e-4, W_a=0.3, W_x=1, W_d=0,
                  min_dist=0.5, max_dist=20.):
        """Pretrain autoencoder.

        Parameters
        ----------
        x :
            input features.
        x_raw :
            raw input features.
        fname : str
            path to save autoencoder weights.
        scale_factor : list
            scale factor of input features and raw input features.
        epochs : int optional
            number of epochs.
        info_step : int optional
            interval of showing pretraining loss.
        lr : float optional
            learning rate.
        W_a : float optional
            parameter of reconstruction loss.
        W_x : float optional
            parameter of ZINB loss.
        W_d : float optional
            parameter of pairwise distance loss.
        min_dist : float optional
            minimum distance of pairwise distance loss.
        max_dist : float optional
            maximum distance of pairwise distance loss.

        Returns
        -------
        None.

        """
        x = torch.Tensor(x).to(self.device)
        x_raw = torch.Tensor(x_raw).to(self.device)
        scale_factor = torch.tensor(scale_factor).to(self.device)

        self.train()
        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()), lr=lr, amsgrad=True)
        print("Pretraining stage")
        for epoch in range(epochs):
            A_out, z, _, mean, disp, pi = self.forward(self.G_n, x)

            if W_d:
                Dist_loss = torch.mean(dist_loss(z, min_dist, max_dist=max_dist))
            A_rec_loss = torch.mean(F.mse_loss(A_out, torch.Tensor(self.adj).to(self.device)))
            Zinb_loss = self.zinb_loss(x_raw, mean, disp, pi, scale_factor)
            loss = W_a * A_rec_loss + W_x * Zinb_loss
            if W_d:
                loss += W_d * Dist_loss

            if epoch % info_step == 0:
                if W_d:
                    print("Epoch %3d: ZINB Loss: %.8f, MSE Loss: %.8f, Dist Loss: %.8f" % (
                        epoch + 1, Zinb_loss.item(), A_rec_loss.item(), Dist_loss.item()))
                else:
                    print("Epoch %3d: ZINB Loss: %.8f, MSE Loss: %.8f" % (
                        epoch + 1, Zinb_loss.item(), A_rec_loss.item()))

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        torch.save(self.state_dict(), fname)
        print("Pre_train Finish!")

    def fit(self, x, x_raw, y, scale_factor, epochs=300, lr=5e-4, W_a=0.3, W_x=1, W_c=1.5, info_step=1):
        """Pretrain autoencoder.

        Parameters
        ----------
        x :
            input features.
        x_raw :
            raw input features.
        y : list
            true label.
        scale_factor : list
            scale factor of input features and raw input features.
        epochs : int optional
            number of epochs.
        lr : float optional
            learning rate.
        W_a : float optional
            parameter of reconstruction loss.
        W_x : float optional
            parameter of ZINB loss.
        W_c : float optional
            parameter of clustering loss.
        info_step : int optional
            interval of showing pretraining loss.

        Returns
        -------
        None.

        """
        x = torch.Tensor(x).to(self.device)
        x_raw = torch.Tensor(x_raw).to(self.device)
        scale_factor = torch.tensor(scale_factor).to(self.device)

        # Initializing cluster centers with kmeans
        kmeans = KMeans(self.n_clusters, n_init=20)
        enc_h = self.encoder1(self.G_n, x)
        z = self.encoder2(self.G_n, enc_h)
        kmeans.fit_predict(z.detach().cpu().numpy())
        self.mu.data.copy_(torch.tensor(kmeans.cluster_centers_, dtype=torch.float32).to(self.device))

        self.train()
        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()), lr=lr, amsgrad=True)

        aris = []
        P = {}
        Q = {}

        for epoch in tqdm(range(epochs)):
            A_out, _, q, mean, disp, pi = self.forward(self.G_n, x)
            self.q = q
            p = self.target_distribution(q)
            self.y_pred = self.predict()

            # calculate ari score for model selection
            _, _, ari = self.score(y)
            aris.append(ari)
            p_ = {f'epoch{epoch}': p}
            q_ = {f'epoch{epoch}': q}
            P = {**P, **p_}
            Q = {**Q, **q_}
            print("Epoch %3d, ARI: %.4f, Best ARI: %.4f" % (epoch + 1, ari, max(aris)))

            A_rec_loss = torch.mean(F.mse_loss(A_out, torch.Tensor(self.adj).to(self.device)))
            Zinb_loss = self.zinb_loss(x_raw, mean, disp, pi, scale_factor)
            Cluster_loss = torch.mean(F.kl_div(torch.Tensor(self.y_pred).to(self.device), torch.Tensor(y).to(self.device)))
            loss = W_a * A_rec_loss + W_x * Zinb_loss + W_c * Cluster_loss

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        index = np.argmax(aris)
        self.q = Q[f'epoch{index}']

    def predict(self):
        """Get predictions from the trained model.

        Parameters
        ----------
        None.

        Returns
        -------
        y_pred : np.array
            prediction of given clustering method.

        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        return y_pred

    def score(self, y):
        """Evaluate the trained model.

        Parameters
        ----------
        y : list
            true labels.

        Returns
        -------
        acc : float
            accuracy.
        nmi : float
            normalized mutual information.
        ari : float
            adjusted Rand index.
        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        acc = np.round(cluster_acc(y, y_pred), 5)
        nmi = np.round(metrics.normalized_mutual_info_score(y, y_pred), 5)
        ari = np.round(metrics.adjusted_rand_score(y, y_pred), 5)
        return acc, nmi, ari

    def soft_assign(self, z):
        """Soft assign q with z.

        Parameters
        ----------
        z :
            embedding.

        Returns
        -------
        q :
            soft label.

        """
        q = 1.0 / (1.0 + torch.sum((z.unsqueeze(1) - self.mu) ** 2, dim=2) / self.alpha)
        q = q ** ((self.alpha + 1.0) / 2.0)
        q = (q.t() / torch.sum(q, dim=1)).t()
        return q

    def target_distribution(self, q):
        """Calculate auxiliary target distribution p with q.

        Parameters
        ----------
        q :
            soft label.

        Returns
        -------
        p :
            target distribution.

        """
        p = q ** 2 / q.sum(0)
        return (p.t() / p.sum(1)).t()


class DecoderA(nn.Module):
    """Decoder class for adjacency matrix reconstruction.

    Parameters
    ----------
    latent_dim : int optional
        dimension of latent embedding.
    adj_dim : int optional
        dimension of adjacency matrix.
    activation : optional
        activation function.
    dropout : float optional
        dropout rate.

    """

    def __init__(self, latent_dim=15, adj_dim=32, activation=torch.sigmoid, dropout=0):
        super(DecoderA, self).__init__()
        self.dropout = dropout
        self.activation = activation
        self.dec_1 = nn.Linear(latent_dim, adj_dim)

    def forward(self, z):
        """Forward propagation.

        Parameters
        ----------
        z :
            embedding.

        Returns
        -------
        adj :
            reconstructed adjacency matrix.

        """
        dec_h = self.dec_1(z)
        z0 = F.dropout(dec_h, self.dropout)
        adj = self.activation(torch.mm(z0, z0.t()))
        return adj


class DecoderX(nn.Module):
    """scTAG class.

    Parameters
    ----------
    input_dim : int
        dimension of input feature.
    n_z : int
        dimension of latent embedding.
    n_dec_1 : int optional
        number of nodes of decoder layer 1.
    n_dec_2 : int optional
        number of nodes of decoder layer 2.
    n_dec_3 : int optional
        number of nodes of decoder layer 3.

    """

    def __init__(self, input_dim, n_z, n_dec_1=128, n_dec_2=256, n_dec_3=512):
        super(DecoderX, self).__init__()
        self.n_dec_3 = n_dec_3
        self.input_dim = input_dim
        self.dec_1 = nn.Linear(n_z, n_dec_1)
        self.dec_2 = nn.Linear(n_dec_1, n_dec_2)
        self.dec_3 = nn.Linear(n_dec_2, n_dec_3)
        self.dec_mean = nn.Sequential(nn.Linear(self.n_dec_3, self.input_dim), MeanAct())
        self.dec_disp = nn.Sequential(nn.Linear(self.n_dec_3, self.input_dim), DispAct())
        self.dec_pi = nn.Sequential(nn.Linear(self.n_dec_3, self.input_dim), nn.Sigmoid())

    def forward(self, z):
        """Forward propagation.

        Parameters
        ----------
        z :
            embedding.

        Returns
        -------
        _mean :
            data mean from ZINB.
        _disp :
            data dispersion from ZINB.
        _pi :
            data dropout probability from ZINB.

        """
        dec_h1 = F.relu(self.dec_1(z))
        dec_h2 = F.relu(self.dec_2(dec_h1))
        dec_h3 = F.relu(self.dec_3(dec_h2))
        _mean = self.dec_mean(dec_h3)
        _disp = self.dec_disp(dec_h3)
        _pi = self.dec_pi(dec_h3)
        return _mean, _disp, _pi


class MeanAct(nn.Module):
    """Mean activation class.
    """

    def __init__(self):
        super(MeanAct, self).__init__()

    def forward(self, x):
        return torch.clamp(torch.exp(x), min=1e-5, max=1e6)


class DispAct(nn.Module):
    """Dispersion activation class.
    """

    def __init__(self):
        super(DispAct, self).__init__()

    def forward(self, x):
        return torch.clamp(F.softplus(x), min=1e-4, max=1e4)


"""
This is the implementation of scDeepCluster.
Extended from https://github.com/ttgump/scDeepCluster

Reference
----------
Tian, Tian, et al. "Clustering single-cell RNA-seq data with a model-based deep learning approach." Nature Machine Intelligence 1.4 (2019): 191-198.
"""

import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.nn import Parameter
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset
import numpy as np
from sklearn.cluster import KMeans
import math, os
from sklearn import metrics

from dance.utils.loss import ZINBLoss
from dance.utils.metrics import cluster_acc


def buildNetwork(layers, type, activation="relu"):
    """Build network layer.

    Parameters
    ----------
    layers : list
        dimensions of layers.
    type : str
        type of network.
    activation : str optional
        activation function.

    Returns
    -------
    net :
        torch.nn network.

    """
    net = []
    for i in range(1, len(layers)):
        net.append(nn.Linear(layers[i - 1], layers[i]))
        if activation == "relu":
            net.append(nn.ReLU())
        elif activation == "sigmoid":
            net.append(nn.Sigmoid())
    net = nn.Sequential(*net)
    return net


def euclidean_dist(x, y):
    """Calculate Euclidean distance between x and y.
    """
    return torch.sum(torch.square(x - y), dim=1)


class ScDeepCluster(nn.Module):
    """scDeepCluster class.

    Parameters
    ----------
    input_dim : int
        dimension of encoder input.
    z_dim : int
        dimension of embedding.
    encodeLayer : list optional
        dimensions of encoder layers.
    decodeLayer : list optional
        dimensions of decoder layers.
    activation : str optional
        activation function.
    sigma : float optional
        parameter of Gaussian noise.
    alpha : float optional
        parameter of soft assign.
    gamma : float optional
        parameter of cluster loss.
    device : str optional
        computing device.

    """

    def __init__(self, input_dim, z_dim, encodeLayer=[], decodeLayer=[],
                 activation="relu", sigma=1., alpha=1., gamma=1., device="cuda"):
        super(ScDeepCluster, self).__init__()
        self.z_dim = z_dim
        self.activation = activation
        self.sigma = sigma
        self.alpha = alpha
        self.gamma = gamma
        self.device = device
        self.encoder = buildNetwork([input_dim] + encodeLayer, type="encode", activation=activation)
        self.decoder = buildNetwork([z_dim] + decodeLayer, type="decode", activation=activation)
        self._enc_mu = nn.Linear(encodeLayer[-1], z_dim)
        self._dec_mean = nn.Sequential(nn.Linear(decodeLayer[-1], input_dim), MeanAct())
        self._dec_disp = nn.Sequential(nn.Linear(decodeLayer[-1], input_dim), DispAct())
        self._dec_pi = nn.Sequential(nn.Linear(decodeLayer[-1], input_dim), nn.Sigmoid())

        self.zinb_loss = ZINBLoss().to(self.device)
        self.to(device)

    def save_model(self, path):
        """Save model to path.

        Parameters
        ----------
        path : str
            path to save model.

        Returns
        -------
        None.

        """
        torch.save(self.state_dict(), path)

    def load_model(self, path):
        """Load model from path.

        Parameters
        ----------
        path : str
            path to load model.

        Returns
        -------
        None.

        """
        pretrained_dict = torch.load(path, map_location=lambda storage, loc: storage)
        model_dict = self.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
        model_dict.update(pretrained_dict)
        self.load_state_dict(model_dict)

    def soft_assign(self, z):
        """Soft assign q with z.

        Parameters
        ----------
        z :
            embedding.

        Returns
        -------
        q :
            soft label.

        """
        q = 1.0 / (1.0 + torch.sum((z.unsqueeze(1) - self.mu) ** 2, dim=2) / self.alpha)
        q = q ** ((self.alpha + 1.0) / 2.0)
        q = (q.t() / torch.sum(q, dim=1)).t()
        return q

    def target_distribution(self, q):
        """Calculate auxiliary target distribution p with q.

        Parameters
        ----------
        q :
            soft label.

        Returns
        -------
        p :
            target distribution.

        """
        p = q ** 2 / q.sum(0)
        return (p.t() / p.sum(1)).t()

    def forwardAE(self, x):
        """Forward propagation of autoencoder.

        Parameters
        ----------
        x :
            input features.

        Returns
        -------
        z0 :
            embedding.
        _mean :
            data mean from ZINB.
        _disp :
            data dispersion from ZINB.
        _pi :
            data dropout probability from ZINB.

        """
        h = self.encoder(x + torch.randn_like(x) * self.sigma)
        z = self._enc_mu(h)
        h = self.decoder(z)
        _mean = self._dec_mean(h)
        _disp = self._dec_disp(h)
        _pi = self._dec_pi(h)

        h0 = self.encoder(x)
        z0 = self._enc_mu(h0)
        return z0, _mean, _disp, _pi

    def forward(self, x):
        """Forward propagation.

        Parameters
        ----------
        x :
            input features.

        Returns
        -------
        z0 :
            embedding.
        q :
            soft label.
        _mean :
            data mean from ZINB.
        _disp :
            data dispersion from ZINB.
        _pi :
            data dropout probability from ZINB.

        """
        h = self.encoder(x + torch.randn_like(x) * self.sigma)
        z = self._enc_mu(h)
        h = self.decoder(z)
        _mean = self._dec_mean(h)
        _disp = self._dec_disp(h)
        _pi = self._dec_pi(h)

        h0 = self.encoder(x)
        z0 = self._enc_mu(h0)
        q = self.soft_assign(z0)
        return z0, q, _mean, _disp, _pi

    def encodeBatch(self, X, batch_size=256):
        """Batch encoder.

        Parameters
        ----------
        X :
            input features.
        batch_size : int optional
            size of batch.

        Returns
        -------
        encoded :
            embedding.

        """
        self.eval()
        encoded = []
        num = X.shape[0]
        num_batch = int(math.ceil(1.0 * X.shape[0] / batch_size))
        for batch_idx in range(num_batch):
            xbatch = X[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
            inputs = Variable(xbatch).to(self.device)
            z, _, _, _ = self.forwardAE(inputs)
            encoded.append(z.data)

        encoded = torch.cat(encoded, dim=0)
        return encoded.to(self.device)

    def cluster_loss(self, p, q):
        """Calculate cluster loss.

        Parameters
        ----------
        p :
            target distribution.
        q :
            soft label.

        Returns
        -------
        loss :
            cluster loss.

        """

        def kld(target, pred):
            return torch.mean(torch.sum(target * torch.log(target / (pred + 1e-6)), dim=-1))

        kldloss = kld(p, q)
        return self.gamma * kldloss

    def pretrain_autoencoder(self, X, X_raw, size_factor, batch_size=256, lr=0.001, epochs=400, ae_save=True,
                             ae_weights='AE_weights.pth.tar'):
        """Pretrain autoencoder.

        Parameters
        ----------
        X :
            input features.
        X_raw :
            raw input features.
        size_factor : float
            size factor of input features and raw input features.
        batch_size : int optional
            size of batch.
        lr : float optional
            learning rate.
        epochs : int optional
            number of epochs.
        ae_save : bool optional
            save autoencoder weights or not.
        ae_weights : str optional
            path to save autoencoder weights.

        Returns
        -------
        None.

        """
        self.train()
        dataset = TensorDataset(torch.Tensor(X), torch.Tensor(X_raw), torch.Tensor(size_factor))
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
        print("Pretraining stage")
        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()), lr=lr, amsgrad=True)
        for epoch in range(epochs):
            loss_val = 0
            for batch_idx, (x_batch, x_raw_batch, sf_batch) in enumerate(dataloader):
                x_tensor = Variable(x_batch).to(self.device)
                x_raw_tensor = Variable(x_raw_batch).to(self.device)
                sf_tensor = Variable(sf_batch).to(self.device)
                _, mean_tensor, disp_tensor, pi_tensor = self.forwardAE(x_tensor)
                loss = self.zinb_loss(x=x_raw_tensor, mean=mean_tensor, disp=disp_tensor, pi=pi_tensor,
                                      scale_factor=sf_tensor)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                loss_val += loss.item() * len(x_batch)
            print('Pretrain epoch %3d, ZINB loss: %.8f' % (epoch + 1, loss_val / X.shape[0]))

        if ae_save:
            torch.save({'ae_state_dict': self.state_dict(),
                        'optimizer_state_dict': optimizer.state_dict()}, ae_weights)

    def save_checkpoint(self, state, index, filename):
        """Save training checkpoint.

        Parameters
        ----------
        state :
            model state
        index : int
            checkpoint index
        filename : str
            filename to save

        Returns
        -------
        None.

        """
        newfilename = os.path.join(filename, 'FTcheckpoint_%d.pth.tar' % index)
        torch.save(state, newfilename)

    def fit(self, X, X_raw, size_factor, n_clusters, init_centroid=None, y=None, y_pred_init=None, lr=1.,
            batch_size=256,
            num_epochs=10, update_interval=1, tol=1e-3, save_dir=""):
        """Train model.

        Parameters
        ----------
        X :
            input features.
        X_raw :
            raw input features.
        size_factor : list
            size factor of input features and raw input features.
        n_clusters : int
            number of clusters.
        init_centroid : list optional
            initialization of centroids. If None, perform kmeans to initialize cluster centers.
        y : list optional
            true label. Used for model selection.
        y_pred_init : list optional
            predicted label for initialization.
        lr : float optional
            learning rate.
        batch_size : int optional
            size of batch.
        num_epochs : int optional
            number of epochs.
        update_interval : int optional
            update interval of soft label and target distribution.
        tol : float optional
            tolerance for training loss.
        save_dir : str optional
            path to save model weights.

        Returns
        -------
        None.

        """
        self.train()
        print("Clustering stage")
        X = torch.tensor(X, dtype=torch.float32)
        X_raw = torch.tensor(X_raw, dtype=torch.float32)
        size_factor = torch.tensor(size_factor, dtype=torch.float32)
        self.mu = Parameter(torch.Tensor(n_clusters, self.z_dim).to(self.device))
        optimizer = optim.Adadelta(filter(lambda p: p.requires_grad, self.parameters()), lr=lr, rho=.95)

        print("Initializing cluster centers with kmeans.")
        if init_centroid is None:
            kmeans = KMeans(n_clusters, n_init=20)
            data = self.encodeBatch(X)
            self.y_pred = kmeans.fit_predict(data.data.cpu().numpy())
            self.y_pred_last = self.y_pred
            self.mu.data.copy_(torch.tensor(kmeans.cluster_centers_, dtype=torch.float32))
        else:
            self.mu.data.copy_(torch.tensor(init_centroid, dtype=torch.float32))
            self.y_pred = y_pred_init
            self.y_pred_last = self.y_pred

        num = X.shape[0]
        num_batch = int(math.ceil(1.0 * X.shape[0] / batch_size))

        aris = []
        P = {}
        Q = {}

        for epoch in range(num_epochs):
            if epoch % update_interval == 0:
                # update the targe distribution p
                latent = self.encodeBatch(X.to(self.device))
                q = self.soft_assign(latent)
                self.q = q
                p = self.target_distribution(q).data
                self.y_pred = self.predict()

                # save current model
                if (epoch > 0 and delta_label < tol) or epoch % 10 == 0:
                    self.save_checkpoint({'epoch': epoch + 1,
                                          'state_dict': self.state_dict(),
                                          'mu': self.mu,
                                          'y_pred': self.y_pred,
                                          'y_pred_last': self.y_pred_last,
                                          'y': y
                                          }, epoch + 1, filename=save_dir)
                p_ = {f'epoch{epoch}': p}
                q_ = {f'epoch{epoch}': q}
                P = {**P, **p_}
                Q = {**Q, **q_}

                # check stop criterion
                delta_label = np.sum(self.y_pred != self.y_pred_last).astype(np.float32) / num
                self.y_pred_last = self.y_pred
                if epoch > 0 and delta_label < tol:
                    print('delta_label ', delta_label, '< tol ', tol)
                    print("Reach tolerance threshold. Stopping training.")
                    break

                # calculate ari score for model selection
                _, _, ari = self.score(y)
                aris.append(ari)

            # train 1 epoch for clustering loss
            train_loss = 0.0
            recon_loss_val = 0.0
            cluster_loss_val = 0.0
            for batch_idx in range(num_batch):
                xbatch = X[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                xrawbatch = X_raw[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                sfbatch = size_factor[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                pbatch = p[batch_idx * batch_size: min((batch_idx + 1) * batch_size, num)]
                optimizer.zero_grad()
                inputs = Variable(xbatch).to(self.device)
                rawinputs = Variable(xrawbatch).to(self.device)
                sfinputs = Variable(sfbatch).to(self.device)
                target = Variable(pbatch).to(self.device)

                zbatch, qbatch, meanbatch, dispbatch, pibatch = self.forward(inputs)

                cluster_loss = self.cluster_loss(target, qbatch)
                recon_loss = self.zinb_loss(rawinputs, meanbatch, dispbatch, pibatch, sfinputs)

                loss = cluster_loss * self.gamma + recon_loss
                loss.backward()
                optimizer.step()
                cluster_loss_val += cluster_loss.item() * len(inputs)
                recon_loss_val += recon_loss.item() * len(inputs)
                train_loss += loss.item() * len(inputs)

            print("Epoch %3d: Total: %.8f, Clustering Loss: %.8f, ZINB Loss: %.8f" % (
                epoch + 1, train_loss / num, cluster_loss_val / num, recon_loss_val / num))

        index = update_interval * np.argmax(aris)
        self.q = Q[f'epoch{index}']

    def predict(self):
        """Get predictions from the trained model.

        Parameters
        ----------
        None.

        Returns
        -------
        y_pred : np.array
            prediction of given clustering method.

        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        return y_pred

    def score(self, y):
        """Evaluate the trained model.

        Parameters
        ----------
        y : list
            true labels.

        Returns
        -------
        acc : float
            accuracy.
        nmi : float
            normalized mutual information.
        ari : float
            adjusted Rand index.
        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        acc = np.round(cluster_acc(y, y_pred), 5)
        nmi = np.round(metrics.normalized_mutual_info_score(y, y_pred), 5)
        ari = np.round(metrics.adjusted_rand_score(y, y_pred), 5)
        return acc, nmi, ari


class MeanAct(nn.Module):
    """Mean activation class.
    """

    def __init__(self):
        super(MeanAct, self).__init__()

    def forward(self, x):
        return torch.clamp(torch.exp(x), min=1e-5, max=1e6)


class DispAct(nn.Module):
    """Dispersion activation class.
    """

    def __init__(self):
        super(DispAct, self).__init__()

    def forward(self, x):
        return torch.clamp(F.softplus(x), min=1e-4, max=1e4)


import torch
import torch.utils.data
from torch import nn
from torch.nn import functional as F
import numpy as np
import math
from torch.autograd import Variable
from torch.distributions import Normal, kl_divergence as kl
from sklearn.mixture import GaussianMixture
import torch.utils.data as data_utils
import time
import scipy.stats as stats
from dance.utils.loss import GMM_loss
import collections
from torch.autograd import Variable
from torch import optim
from sklearn.cluster import KMeans
from sklearn import metrics
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.metrics import adjusted_rand_score
from tqdm import trange
import os
import copy

import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)

def save_checkpoint( model, fileName = './saved_model/model_best.pth.tar'):
    folder = os.path.dirname(fileName)
    os.makedirs(folder, exist_ok=True)
    torch.save( model.state_dict(), fileName)


def load_checkpoint(file_path, model, device):

	model.load_state_dict( torch.load(file_path) )
	model.to(device)

	return model

def binary_cross_entropy(recon_x, x):
    return - torch.sum(x * torch.log(recon_x + 1e-8) + (1 - x) * torch.log(1 - recon_x + 1e-8), dim=1)

def log_zinb_positive(x, mu, theta, pi, eps=1e-8):
    x = x.float()

    if theta.ndimension() == 1:
        theta = theta.view(1, theta.size(0))

    softplus_pi = F.softplus(-pi)

    log_theta_eps = torch.log(theta + eps)

    log_theta_mu_eps = torch.log(theta + mu + eps)

    pi_theta_log = -pi + theta * (log_theta_eps - log_theta_mu_eps)

    case_zero = F.softplus(pi_theta_log) - softplus_pi
    mul_case_zero = torch.mul((x < eps).type(torch.float32), case_zero)

    case_non_zero = (-softplus_pi + pi_theta_log
                     + x * (torch.log(mu + eps) - log_theta_mu_eps)
                     + torch.lgamma(x + theta)
                     - torch.lgamma(theta)
                     - torch.lgamma(x + 1))

    mul_case_non_zero = torch.mul((x > eps).type(torch.float32), case_non_zero)

    res = mul_case_zero + mul_case_non_zero

    return - torch.sum(res, dim=1)


def NB_loss(y_true, y_pred, theta, eps=1e-10):
    y_true = y_true.float()
    y_pred = y_pred.float()

    t1 = torch.lgamma(theta + eps) + torch.lgamma(y_true + 1.0) - torch.lgamma(y_true + theta + eps)
    t2 = (theta + y_true) * torch.log(1.0 + (y_pred / (theta + eps))) + (
                y_true * (torch.log(theta + eps) - torch.log(y_pred + eps)))

    final = t1 + t2

    return - torch.sum(final, dim=1)


def mse_loss(y_true, y_pred):
    mask = torch.sign(y_true)

    y_pred = y_pred.float()
    y_true = y_true.float()

    ret = torch.pow((y_pred - y_true) * mask, 2)

    return torch.sum(ret, dim=1)


def poisson_loss(y_true, y_pred):
    y_pred = y_pred.float()
    y_true = y_true.float()

    ret = y_pred - y_true * torch.log(y_pred + 1e-10) + torch.lgamma(y_true + 1.0)

    return torch.sum(ret, dim=1)

def adjust_learning_rate(init_lr, optimizer, iteration, max_lr, adjust_epoch):

	lr = max(init_lr * (0.9 ** (iteration//adjust_epoch)), max_lr)
	for param_group in optimizer.param_groups:
		param_group["lr"] = lr

	return lr

def build_multi_layers(layers, use_batch_norm=True, dropout_rate=0.1):
    """Build multilayer linear perceptron"""
    if dropout_rate > 0:
        fc_layers = nn.Sequential(
            collections.OrderedDict(
                [
                    (
                        "Layer {}".format(i),
                        nn.Sequential(
                            nn.Linear(n_in, n_out),
                            nn.BatchNorm1d(n_out, momentum=0.01, eps=0.001),
                            nn.ReLU(),
                            nn.Dropout(p=dropout_rate),
                        ),
                    )

                    for i, (n_in, n_out) in enumerate(zip(layers[:-1], layers[1:]))
                ]
            )
        )

    else:
        fc_layers = nn.Sequential(
            collections.OrderedDict(
                [
                    (
                        "Layer {}".format(i),
                        nn.Sequential(
                            nn.Linear(n_in, n_out),
                            nn.BatchNorm1d(n_out, momentum=0.01, eps=0.001),
                            nn.ReLU(),
                        ),
                    )

                    for i, (n_in, n_out) in enumerate(zip(layers[:-1], layers[1:]))
                ]
            )
        )

    return fc_layers


class Encoder(nn.Module):
    ## for one modulity
    def __init__(self, layer, hidden, Z_DIMS, dropout_rate=0.1):
        super(Encoder, self).__init__()

        if len(layer) > 1:
            self.fc1 = build_multi_layers(layers=layer, dropout_rate=dropout_rate)

        self.layer = layer
        self.fc_means = nn.Linear(hidden, Z_DIMS)
        self.fc_logvar = nn.Linear(hidden, Z_DIMS)

    def reparametrize(self, means, logvar):

        if self.training:
            std = logvar.mul(0.5).exp_()
            eps = Variable(std.data.new(std.size()).normal_())
            return eps.mul(std).add_(means)
        else:
            return means

    def forward(self, x):

        if len(self.layer) > 1:
            h = self.fc1(x)
        else:
            h = x
        mean_x = self.fc_means(h)
        logvar_x = self.fc_logvar(h)
        latent = self.reparametrize(mean_x, logvar_x)

        return mean_x, logvar_x, latent


class DecoderZINB(nn.Module):
    ### for scRNA-seq

    def __init__(self, layer, hidden, input_size, dropout_rate=0.1):

        super(DecoderZINB, self).__init__()

        if len(layer) > 1:
            self.decoder = build_multi_layers(layer, dropout_rate=dropout_rate)

        self.decoder_scale = nn.Linear(hidden, input_size)
        self.decoder_r = nn.Linear(hidden, input_size)
        self.dropout = nn.Linear(hidden, input_size)

        self.layer = layer

    def forward(self, z, library):

        if len(self.layer) > 1:
            latent = self.decoder(z)
        else:
            latent = z

        normalized_x = F.softmax(self.decoder_scale(latent), dim=1)  ## mean gamma

        recon_final = torch.exp(library) * normalized_x  ##mu
        disper_x = self.decoder_r(latent)  ### theta
        disper_x = torch.exp(disper_x)
        dropout_rate = self.dropout(latent)

        return dict(normalized=normalized_x,
                    disperation=disper_x,
                    imputation=recon_final,
                    dropoutrate=dropout_rate
                    )


class DecoderNB(nn.Module):

    ### for scRNA-seq

    def __init__(self, layer, hidden, input_size):
        super(DecoderNB, self).__init__()

        self.decoder = build_multi_layers(layers=layer)

        self.decoder_scale = nn.Linear(hidden, input_size)
        self.decoder_r = nn.Linear(hidden, input_size)

    def forward(self, z, library):
        latent = self.decoder(z)

        normalized_x = F.softmax(self.decoder_scale(latent), dim=1)  ## mean gamma

        recon_final = torch.exp(library) * normalized_x  ##mu
        disper_x = self.decoder_r(latent)  ### theta
        disper_x = torch.exp(disper_x)

        return dict(normalized=normalized_x,
                    disperation=disper_x,
                    imputation=recon_final
                    )


class Decoder(nn.Module):
    ### for scATAC-seq
    def __init__(self, layer, hidden, input_size, Type="Bernoulli", dropout_rate=0.1):
        super(Decoder, self).__init__()

        if len(layer) > 1:
            self.decoder = build_multi_layers(layer, dropout_rate=dropout_rate)

        self.decoder_x = nn.Linear(hidden, input_size)
        self.Type = Type
        self.layer = layer

    def forward(self, z):

        if len(self.layer) > 1:
            latent = self.decoder(z)
        else:
            latent = z

        recon_x = self.decoder_x(latent)

        if self.Type == "Bernoulli":
            Final_x = torch.sigmoid(recon_x)

        elif self.Type == "Gaussian":
            Final_x = F.softmax(recon_x, dim=1)

        elif self.Type == "Gaussian1":
            Final_x = torch.sigmoid(recon_x)

        else:
            Final_x = F.relu(recon_x)

        return Final_x

class scMVAE(nn.Module):
    ## scMVAE-PoE

    def __init__(self, encoder_1, hidden_1, Z_DIMS, decoder_share, share_hidden,
                 decoder_1, hidden_2, encoder_l, hidden3, encoder_2, hidden_4,
                 encoder_l1, hidden3_1, decoder_2, hidden_5, drop_rate,
                 log_variational=True, Type='Bernoulli', device='cpu',
                 n_centroids=19, penality="GMM", model=2
                 ):

        super(scMVAE, self).__init__()

        self.X1_encoder = Encoder(encoder_1, hidden_1, Z_DIMS, dropout_rate=drop_rate)
        self.X1_encoder_l = Encoder(encoder_l, hidden3, 1, dropout_rate=drop_rate)

        self.X1_decoder = DecoderZINB(decoder_1, hidden_2, encoder_1[0], dropout_rate=drop_rate)

        self.X2_encoder = Encoder(encoder_2, hidden_4, Z_DIMS, dropout_rate=drop_rate)

        self.decode_share = build_multi_layers(decoder_share, dropout_rate=drop_rate)

        if Type == 'ZINB':
            self.X2_encoder_l = Encoder(encoder_l1, hidden3_1, 1, dropout_rate=drop_rate)
            self.decoder_x2 = DecoderZINB(decoder_2, hidden_5, encoder_2[0],
                                          dropout_rate=drop_rate)
        elif Type == 'Bernoulli':
            self.decoder_x2 = Decoder(decoder_2, hidden_5, encoder_2[0], Type,
                                      dropout_rate=drop_rate)
        elif Type == "Possion":
            self.decoder_x2 = Decoder(decoder_2, hidden_5, encoder_2[0], Type,
                                      dropout_rate=drop_rate)
        else:
            self.decoder_x2 = Decoder(decoder_2, hidden_5, encoder_2[0], Type,
                                      dropout_rate=drop_rate)

        self.experts = ProductOfExperts()
        self.Z_DIMS = Z_DIMS
        self.share_hidden = share_hidden
        self.log_variational = log_variational
        self.Type = Type
        self.decoder_share = decoder_share
        self.decoder_1 = decoder_1
        self.n_centroids = n_centroids
        self.penality = penality
        self.device = device
        self.model = model

        self.pi = nn.Parameter(torch.ones(n_centroids) / n_centroids)  # pc
        self.mu_c = nn.Parameter(torch.zeros(Z_DIMS, n_centroids))  # mu
        self.var_c = nn.Parameter(torch.ones(Z_DIMS, n_centroids))  # sigma^2

    def _reparametrize(self, means, logvar):

        if self.training:
            std = logvar.mul(0.5).exp_()
            eps = Variable(std.data.new(std.size()).normal_())
            return eps.mul(std).add_(means)
        else:
            return means

    def _encode_modalities(self, X1=None, X2=None):

        if X1 is not None:
            batch_size = X1.size(0)
        else:
            batch_size = X2.size(0)

        # Initialization
        means, logvar = prior_expert((1, batch_size, self.Z_DIMS))

        means = means.to(self.device)
        logvar = logvar.to(self.device)

        # Support for weak supervision setting
        if X1 is not None:
            X1_mean, X1_logvar, _ = self.X1_encoder(X1)
            means = torch.cat((means, X1_mean.unsqueeze(0)), dim=0)
            logvar = torch.cat((logvar, X1_logvar.unsqueeze(0)), dim=0)

        if X2 is not None:
            X2_mean, X2_logvar, _ = self.X2_encoder(X2)
            means = torch.cat((means, X2_mean.unsqueeze(0)), dim=0)
            logvar = torch.cat((logvar, X2_logvar.unsqueeze(0)), dim=0)

        # Combine the gaussians
        means, logvar = self.experts(means, logvar)
        return means, logvar

    def _inference(self, X1=None, X2=None):

        X1_ = X1
        X2_ = X2
        ### X1 processing
        mean_l, logvar_l, library = None, None, None

        if X1 is not None:
            if self.log_variational:
                X1_ = torch.log(X1_ + 1)

            mean_l, logvar_l, library = self.X1_encoder_l(X1_)

        ### X2 processing
        mean_l2, logvar_l2, library2 = None, None, None

        if X2 is not None:

            if self.Type == 'ZINB':
                if self.log_variational:
                    X2_ = torch.log(X2_ + 1)
                    mean_l2, logvar_l2, library2 = self.X2_encoder_l(X2_)

        means, logvar = self._encode_modalities(X1_, X2_)

        z = self._reparametrize(means, logvar)

        if len(self.decoder_share) > 1:
            latents = self.decode_share(z)

            if self.model == 0:
                latent_1 = latents
                latent_2 = latents

            elif self.model == 1:
                latent_1 = latents[:, :self.share_hidden]
                latent_2 = latents[:, self.share_hidden:]

            elif self.model == 2:
                latent_1 = torch.cat((z, latents[:, :self.share_hidden]), 1)
                latent_2 = latents[:, self.share_hidden:]

            else:
                latent_1 = torch.cat((z, latents), 1)
                latent_2 = latents
        else:
            latent_1 = z
            latent_2 = z

        # Reconstruct
        output = self.X1_decoder(latent_1, library)

        normalized_x = output["normalized"]
        recon_X1 = output["imputation"]
        disper_x = output["disperation"]
        dropout_rate = output["dropoutrate"]

        if self.Type == 'ZINB':
            results = self.decoder_x2(latent_2, library2)

            norma_x2 = results["normalized"]
            recon_X2 = results["imputation"]
            disper_x2 = results["disperation"]
            dropout_rate_2 = results["dropoutrate"]

        else:
            recon_X2 = self.decoder_x2(latent_2)
            norma_x2, disper_x2, dropout_rate_2 = None, None, None

        return dict(norm_x1=normalized_x, disper_x=disper_x,
                    recon_x1=recon_X1, dropout_rate=dropout_rate,
                    norm_x2=norma_x2, disper_x2=disper_x2,
                    recon_x_2=recon_X2, dropout_rate_2=dropout_rate_2,
                    mean_l=mean_l, logvar_l=logvar_l,
                    library=library, mean_l2=mean_l2,
                    logvar_l2=logvar_l2, library2=library2,
                    mean_z=means, logvar_z=logvar,
                    latent_z=z,
                    )

    def forward(self, X1, X2, local_l_mean, local_l_var, local_l_mean1, local_l_var1):
        """Forward function for torch.nn.Module. An alias of encode_Batch function.

        Parameters
        ----------
        X1 : torch.utils.data.DataLoader
            Dataloader for dataset.
        X2 :
        local_l_mean:
        local_l_var:

        Returns
        -------
        latent_z1 : numpy.ndarray
            Latent representation of modality 1.
        latent_z2 : numpy.ndarray
            Latent representation of modality 2.
        norm_x1 : numpy.ndarray
            Normalized representation of modality 1.
        recon_x1 : numpy.ndarray
            Reconstruction result of modality 1.
        norm_x2 : numpy.ndarray
            Normalized representation of modality 2.
        recon_x2 : numpy.ndarray
            Reconstruction result of modality 2.
        """
        result = self._inference(X1, X2)

        disper_x = result["disper_x"]
        recon_x1 = result["recon_x1"]
        dropout_rate = result["dropout_rate"]

        disper_x2 = result["disper_x2"]
        recon_x_2 = result["recon_x_2"]
        dropout_rate_2 = result["dropout_rate_2"]

        if X1 is not None:
            mean_l = result["mean_l"]
            logvar_l = result["logvar_l"]

            kl_divergence_l = kl(Normal(mean_l, torch.exp(logvar_l)),
                                 Normal(local_l_mean, torch.sqrt(local_l_var))).sum(dim=1)
        else:
            kl_divergence_l = torch.tensor(0.0)

        if X2 is not None:
            if self.Type == 'ZINB':
                mean_l2 = result["mean_l2"]
                logvar_l2 = result["library2"]
                kl_divergence_l2 = kl(Normal(mean_l2, torch.exp(logvar_l2)),
                                      Normal(local_l_mean, torch.sqrt(local_l_var))).sum(dim=1)
            else:
                kl_divergence_l2 = torch.tensor(0.0)
        else:
            kl_divergence_l2 = torch.tensor(0.0)

        mean_z = result["mean_z"]
        logvar_z = result["logvar_z"]
        latent_z = result["latent_z"]

        if self.penality == "GMM":
            gamma, mu_c, var_c, pi = self._get_gamma(latent_z)  # , self.n_centroids, c_params)
            kl_divergence_z = GMM_loss(gamma, (mu_c, var_c, pi), (mean_z, torch.exp(logvar_z)))

        else:
            mean = torch.zeros_like(mean_z)
            scale = torch.ones_like(logvar_z)
            kl_divergence_z = kl(Normal(mean_z, torch.exp(logvar_z)),
                                 Normal(mean, scale)).sum(dim=1)

        loss1, loss2 = get_both_recon_loss(X1, recon_x1, disper_x, dropout_rate,
                                           X2, recon_x_2, disper_x2, dropout_rate_2,
                                           "ZINB", self.Type)

        return loss1, loss2, kl_divergence_l, kl_divergence_l2, kl_divergence_z

    def _out_Batch(self, Dataloader, out='Z', transforms=None):
        output = []

        for i, (X1, X2) in enumerate(Dataloader):

            X1 = X1.view(X1.size(0), -1).float().to(self.device)
            X2 = X2.view(X2.size(0), -1).float().to(self.device)

            result = self._inference(X1, X2)

            if out == 'Z':
                output.append(result["latent_z"].detach().cpu())
            elif out == 'recon_X1':
                output.append(result["recon_x1"].detach().cpu().data)
            elif out == 'Norm_X1':
                output.append(result["norm_x1"].detach().cpu().data)
            elif out == 'recon_X2':
                output.append(result["recon_x_2"].detach().cpu().data)

            elif out == 'logit':
                print('Logit not supported.')
                # output.append(self.get_gamma(z)[0].cpu().detach())

        output = torch.cat(output).numpy()

        return output

    def _get_gamma(self, z):

        n_centroids = self.n_centroids

        N = z.size(0)
        z = z.unsqueeze(2).expand(z.size(0), z.size(1), n_centroids)
        pi = self.pi.repeat(N, 1)  # NxK
        mu_c = self.mu_c.repeat(N, 1, 1)  # NxDxK
        var_c = self.var_c.repeat(N, 1, 1)  # NxDxK

        # p(c,z) = p(c)*p(z|c) as p_c_z
        p_c_z = torch.exp(
            torch.log(pi) - torch.sum(0.5 * torch.log(2 * math.pi * var_c) + (z - mu_c) ** 2 / (2 * var_c),
                                      dim=1)) + 1e-10
        gamma = p_c_z / torch.sum(p_c_z, dim=1, keepdim=True)

        return gamma, mu_c, var_c, pi


    def init_gmm_params(self, Dataloader):
        """This function will initialize the parameters for PoE model.

        Parameters
        ----------
        Dataloader : torch.utils.data.DataLoader
            Dataloader for the whole dataset.

        Returns
        -------
        None.
        """
        gmm = GaussianMixture(n_components=self.n_centroids, covariance_type='diag')

        latent_z = self._out_Batch(Dataloader, out='Z')
        gmm.fit(latent_z)

        self.mu_c.data.copy_(torch.from_numpy(gmm.means_.T.astype(np.float32)))
        self.var_c.data.copy_(torch.from_numpy(gmm.covariances_.T.astype(np.float32)))

    def _denoise_batch(self, total_loader):
        # processing large-scale datasets
        latent_z = []
        norm_x1 = []
        recon_x_2 = []
        recon_x1 = []
        norm_x2 = []

        for batch_idx, (X1, X2) in enumerate(total_loader):
            X1 = X1.to(self.device)
            X2 = X2.to(self.device)

            X1 = Variable(X1)
            X2 = Variable(X2)

            result = self._inference(X1, X2)

            latent_z.append(result["latent_z"].data.cpu().numpy())
            recon_x_2.append(result["recon_x_2"].data.cpu().numpy())
            recon_x1.append(result["recon_x1"].data.cpu().numpy())
            norm_x1.append(result["norm_x1"].data.cpu().numpy())
            norm_x2.append(result["norm_x2"].data.cpu().numpy())

        latent_z = np.concatenate(latent_z)
        recon_x_2 = np.concatenate(recon_x_2)
        recon_x1 = np.concatenate(recon_x1)
        norm_x1 = np.concatenate(norm_x1)
        norm_x2 = np.concatenate(norm_x2)

        return latent_z, recon_x1, norm_x1, recon_x_2, norm_x2

    def fit(self, args, train, valid, final_rate, scale_factor, device):
        """fit function for training

        Parameters
        ----------
        train : torch.utils.data.DataLoader
            Dataloader for training dataset.
        valid : torch.utils.data.DataLoader
            Dataloader for testing dataset.
        final_rate : torch.utils.data.DataLoader
            Dataloader for both training and testing dataset, for extra evaluation purpose.
        scale_factor : str
            Type of modality 1.
        device : torch.device

        Returns
        -------
        None.
        """
        train_loader = data_utils.DataLoader(train, batch_size=args.batch_size, shuffle=True)
        test_loader = data_utils.DataLoader(valid, batch_size=len(valid), shuffle=False)

        # args.max_epoch = 500
        # args.epoch_per_test = 10
        train_loss_list = []

        flag_break = 0
        epoch_count = 0
        reco_epoch_test = 0
        test_like_max = 100000

        params = filter(lambda p: p.requires_grad, self.parameters())
        optimizer = optim.Adam(params, lr=args.lr, weight_decay=args.weight_decay, eps=args.eps)

        epoch = 0
        iteration = 0
        start = time.time()

        with trange(args.max_epoch, disable=True) as pbar:
            while True:
                self.train()

                epoch += 1
                epoch_lr = adjust_learning_rate(args.lr, optimizer, epoch, final_rate, 10)
                kl_weight = min(1, epoch / args.anneal_epoch)

                for batch_idx, (X1, lib_m1, lib_v1, lib_m2, lib_v2, X2) in enumerate(train_loader):
                    X1, X2 = X1.float().to(device), X2.float().to(device)
                    lib_m1, lib_v1 = lib_m1.to(device), lib_v1.to(device)
                    lib_m2, lib_v2 = lib_m2.to(device), lib_v2.to(device)

                    X1, X2 = Variable(X1), Variable(X2)
                    lib_m1, lib_v1 = Variable(lib_m1), Variable(lib_v1)
                    lib_m2, lib_v2 = Variable(lib_m2), Variable(lib_v2)

                    optimizer.zero_grad()

                    loss1, loss2, kl_divergence_l, kl_divergence_l1, kl_divergence_z = self(X1.float(), X2.float(),
                                                                                             lib_m1,
                                                                                             lib_v1,
                                                                                             lib_m2,
                                                                                             lib_v2)
                    loss = torch.mean((scale_factor * loss1 + loss2 + kl_divergence_l + kl_divergence_l1) + (
                            kl_weight * (kl_divergence_z)))

                    loss.backward()
                    optimizer.step()

                    iteration += 1

                epoch_count += 1

                if epoch % args.epoch_per_test == 0 and epoch > 0:

                    self.eval()

                    with torch.no_grad():

                        for batch_idx, (X1, lib_m1, lib_v1, lib_m2, lib_v2, X2) in enumerate(test_loader):

                            X1, X2 = X1.float().to(device), X2.float().to(device)
                            lib_v1, lib_m1 = lib_v1.to(device), lib_m1.to(device)
                            lib_v2, lib_m2 = lib_v2.to(device), lib_m2.to(device)

                            X1, X2 = Variable(X1), Variable(X2)
                            lib_m1, lib_v1 = Variable(lib_m1), Variable(lib_v1)
                            lib_m2, lib_v2 = Variable(lib_m2), Variable(lib_v2)

                            loss1, loss2, kl_divergence_l, kl_divergence_l1, kl_divergence_z = self(X1.float(),
                                                                                                     X2.float(),
                                                                                                     lib_m1, lib_v1,
                                                                                                     lib_m2,
                                                                                                     lib_v2)
                            test_loss = torch.mean(
                                (scale_factor * loss1 + loss2 + kl_divergence_l + kl_divergence_l1) + (
                                        kl_weight * (kl_divergence_z)))

                            train_loss_list.append(test_loss.item())

                            if math.isnan(test_loss.item()):
                                flag_break = 1
                                break

                            if test_like_max > test_loss.item():
                                test_like_max = test_loss.item()
                                epoch_count = 0

                                save_checkpoint(self)

                                print(str(epoch) + "   " + str(loss.item()) + "   " + str(test_loss.item()) + "   " +
                                      str(torch.mean(loss1).item()) + "   " + str(torch.mean(loss2).item()) +
                                      "  kl_divergence_l:  " + str(
                                    torch.mean(kl_divergence_l).item()) + " kl_weight: " + str(kl_weight) +
                                      " kl_divergence_z: " + str(torch.mean(kl_divergence_z).item()))


                if flag_break == 1:
                    reco_epoch_test = epoch
                    status = " With NA, training failed. "
                    break

                if epoch >= args.max_epoch:
                    reco_epoch_test = epoch
                    status = f" Reached {args.max_epoch} epoch, training complete. "
                    break

                if len(train_loss_list) >= 2:
                    if abs(train_loss_list[-1] - train_loss_list[-2]) / train_loss_list[-2] < 1e-4:
                        reco_epoch_test = epoch
                        status = " Training for the train dataset is converged! "
                        break

        duration = time.time() - start
        print('Finish training, total time: ' + str(duration) + 's' + " epoch: " + str(
            reco_epoch_test) + " status: " + status)

        load_checkpoint('./saved_model/model_best.pth.tar', self, device)

    def predict(self, X1, X2, out='Z', device='cpu'):
        """Predict function to get prediction.

        Parameters
        ----------
        X1 : torch.Tensor
            Features of modality 1.
        X2 : torch.Tensor
            Features of modality 2.
        out : str optional
            The ground truth labels for evaluation.

        Returns
        -------
        result : torch.Tensor
            The requested result, by default to be embedding in latent space (a.k.a 'Z').
        """
        with torch.no_grad():
            X1 = X1.view(X1.size(0), -1).float()
            X2 = X2.view(X2.size(0), -1).float()
            if device=='cpu':
                model = copy.deepcopy(self).to('cpu')
                model.device = 'cpu'
                X1 = X1.to('cpu')
                X2 = X2.to('cpu')
                result = model._inference(X1, X2)
            else:
                X1 = X1.to(self.device)
                X2 = X2.to(self.device)
                result = self._inference(X1, X2)

            if out == 'Z':
                return result["latent_z"]
            elif out == 'recon_X1':
                return result["recon_x1"]
            elif out == 'Norm_X1':
                return result["norm_x1"]
            elif out == 'recon_X2':
                return result["recon_x_2"]
            elif out == 'logit':
                print('Logit not supported.')
                # output.append(self.get_gamma(z)[0].cpu().detach())
                return None

    def score(self, X1, X2, adata_sol):
        """score function to get score of prediction.

        Parameters
        ----------
        X1 : torch.Tensor
            Features of modality 1.
        X2 : torch.Tensor
            Features of modality 2.
        adata_sol : anndata.Anndata
            The ground truth labels for evaluation.

        Returns
        -------
        NMI_score : float
            NMI eval score.
        ARI_score : float
            ARI eval score.
        """

        emb = self.predict(X1, X2).cpu().numpy()

        kmeans = KMeans(n_clusters=10, n_init=5, random_state=200)

        # adata.obs['batch'] = adata_sol.obs['batch'][adata.obs_names]
        # adata.obs['cell_type'] = adata_sol.obs['cell_type'][adata.obs_names]
        true_labels = adata_sol.obs['cell_type'].to_numpy()

        pred_labels = kmeans.fit_predict(emb)
        NMI_score = round(normalized_mutual_info_score(true_labels, pred_labels, average_method='max'), 3)
        ARI_score = round(adjusted_rand_score(true_labels, pred_labels), 3)

        # print('ARI: ' + str(ARI_score) + ' NMI: ' + str(NMI_score))
        return NMI_score, ARI_score

class ProductOfExperts(nn.Module):
    """Return parameters for product of independent experts.
    See https://arxiv.org/pdf/1410.7827.pdf for equations.
    @param mu: M x D for M experts
    @param logvar: M x D for M experts
    """

    def forward(self, mu, logvar, eps=1e-8):
        var = torch.exp(logvar) + eps
        # precision of i-th Gaussian expert at point x
        T = 1. / var
        pd_mu = torch.sum(mu * T, dim=0) / torch.sum(T, dim=0)
        pd_var = 1. / torch.sum(T, dim=0)
        pd_logvar = torch.log(pd_var)
        return pd_mu, pd_logvar


def prior_expert(size):
    """Universal prior expert. Here we use a spherical
    Gaussian: N(0, 1).
    @param size: integer
                 dimensionality of Gaussian
    """
    mu = Variable(torch.zeros(size))
    logvar = Variable(torch.log(torch.ones(size)))

    return mu, logvar


def get_triple_recon_loss(x1=None, px1_rate=None, px1_r=None, px1_dropout=None,
                          x2=None, px2_rate=None, px2_r=None, px2_dropout=None,
                          px12_rate=None, px12_r=None, px12_dropout=None,
                          Type1="ZINB", Type="Bernoulli"):
    reconst_loss1 = log_zinb_positive(x1, px1_rate, px1_r, px1_dropout)
    reconst_loss12 = binary_cross_entropy(px12_rate, x2)

    if x2 is not None:
        reconst_loss2 = binary_cross_entropy(px2_rate, x2)
    else:
        reconst_loss2 = torch.tensor(0.0)

    return reconst_loss1, reconst_loss2, reconst_loss12


def get_both_recon_loss(x1=None, px1_rate=None, px1_r=None, px1_dropout=None,
                        x2=None, px2_rate=None, px2_r=None, px2_dropout=None,
                        Type1="ZINB", Type="Bernoulli"):
    # Reconstruction Loss
    ## here Type1 for rna-seq, Type for atac-seq
    # reconst_loss1 = log_zinb_positive( x1, px1_rate, px1_r, px1_dropout )
    if x1 is not None:
        reconst_loss1 = log_zinb_positive(x1, px1_rate, px1_r, px1_dropout)
    else:
        reconst_loss1 = torch.tensor(0.0)

    if x2 is not None:

        if Type == "ZINB":
            reconst_loss2 = log_zinb_positive(x2, px2_rate, px2_r, px2_dropout)

        elif Type == 'Bernoulli':
            reconst_loss2 = binary_cross_entropy(px2_rate, x2)

        elif Type == "Possion":
            reconst_loss2 = poisson_loss(x2, px2_rate)

        else:
            reconst_loss2 = mse_loss(x2, px2_rate)
    else:
        reconst_loss2 = torch.tensor(0.0)

    return reconst_loss1, reconst_loss2



import sys
sys.path.append("..")
import warnings
warnings.filterwarnings("ignore")
import argparse
from argparse import Namespace
import numpy as np
import dgl
import random
import torch

from dance.modules.single_modality.clustering.graphsc import *
from dance.transforms.graph_construct import make_graph
from dance.transforms.preprocess import filter_data
from dance.datasets.singlemodality import ClusteringDataset


def pipeline(**args):    
    data = ClusteringDataset(args['data_dir'], args['dataset']).load_data()
    X = data.X
    Y = data.Y
    n_clusters = len(np.unique(Y))

    genes_idx, cells_idx = filter_data(X, highly_genes=args['nb_genes'])
    X = X[cells_idx][:, genes_idx]
    Y = Y[cells_idx]

    t0 = time.time()
    graph = make_graph(
        X,
        Y,
        dense_dim=args['in_feats'],
        node_features=args['node_features'],
        normalize_weights=args['normalize_weights'],
        same_edge_values=args['same_edge_values'],
        edge_norm=args['edge_norm']
    )

    labels = graph.ndata["label"]
    train_ids = np.where(labels != -1)[0]
    sampler = dgl.dataloading.MultiLayerFullNeighborSampler(args['n_layers'])
    dataloader = dgl.dataloading.NodeDataLoader(
        graph,
        train_ids,
        sampler,
        batch_size=args['batch_size'],
        shuffle=True,
        drop_last=False,
        num_workers=args['num_workers']
    )

    t1 = time.time()

    for run in range(args['num_run']):
        t_start = time.time()
        torch.manual_seed(run)
        torch.cuda.manual_seed_all(run)
        np.random.seed(run)
        random.seed(run)

        model = GraphSC(Namespace(**args))
        model.fit(args['epochs'],
                  dataloader,
                  n_clusters,
                  args['learning_rate'],
                  cluster=["KMeans", "Leiden"])
        pred = model.predict(n_clusters, cluster=["KMeans", "Leiden"])
#        print(f'kmeans_pred: {pred.get("kmeans_pred")}\n leiden_pred: {pred.get("leiden_pred")}')
        score = model.score(Y, n_clusters, plot=False, cluster=["KMeans", "Leiden"])
        print(f'kmeans_ari: {score.get("kmeans_ari")}, leiden_ari: {score.get("leiden_ari")}')


parser = argparse.ArgumentParser()
parser.add_argument('-e', '--epochs', default=100, type=int)
parser.add_argument('-cpu', '--use_cpu', default=False, action='store_true')
parser.add_argument('-if', '--in_feats', default=50, type=int)
parser.add_argument('-bs', '--batch_size', default=128, type=int)
parser.add_argument('-nw', '--normalize_weights', default='log_per_cell', choices=['log_per_cell', 'per_cell'])
parser.add_argument('-ac', '--activation', default='relu', choices=['leaky_relu', 'relu', 'prelu', 'gelu'])
parser.add_argument('-drop', '--dropout', default=0.1, type=float)
parser.add_argument('-nf', '--node_features', default='scale', choices=['scale_by_cell', 'scale', 'none'])
parser.add_argument('-sev', '--same_edge_values', default=False, action='store_true')
parser.add_argument('-en', '--edge_norm', default=True, action='store_true')
parser.add_argument('-hr', '--hidden_relu', default=False, action='store_true')
parser.add_argument('-hbn', '--hidden_bn', default=False, action='store_true')
parser.add_argument('-lr', '--learning_rate', type=float, default=1e-5)
parser.add_argument('-nl', '--n_layers', type=int, default=1, choices=[1, 2])
parser.add_argument('-agg', '--agg', default='sum', choices=['sum', 'mean'])
parser.add_argument('-hd', '--hidden_dim', type=int, default=200)
parser.add_argument('-nh', '--n_hidden', type=int, default=1, choices=[0, 1, 2])
parser.add_argument('-h1', '--hidden_1', type=int, default=300)
parser.add_argument('-h2', '--hidden_2', type=int, default=0)
parser.add_argument('-ng', '--nb_genes', type=int, default=3000)
parser.add_argument('-nr', '--num_run', type=int, default=1)
parser.add_argument('-nbw', '--num_workers', type=int, default=0)
parser.add_argument('-eve', '--eval_epoch', default=False, action='store_true')
parser.add_argument('-show', '--show_epoch_ari', default=False, action='store_true')
parser.add_argument('-plot', '--plot', default=False, action='store_true')
parser.add_argument('-dd', '--data_dir', default='./data', type=str)
parser.add_argument('-data', '--dataset', default='10X_PBMC', choices=['10X_PBMC', 'mouse_bladder_cell', 'mouse_ES_cell', 'worm_neuron_cell'])

args = parser.parse_args()
pipeline(**vars(args))

""" Reproduction information
10X PBMC:
python graphsci.py --dataset='10X_PBMC'

Mouse ES:
python graphsci.py --dataset='mouse_ES_cell'

Worm Neuron:
python graphsci.py --dataset='worm_neuron_cell'

Mouse Bladder:
python graphsci.py --dataset='mouse_bladder_cell'
"""


"""
This is the implementation of GrapSCI.
Extended from https://github.com/biomed-AI/GraphSCI

Reference
----------
Rao, Jiahua, et al. "Imputing single-cell RNA-seq data by combining graph convolution and autoencoder neural networks." Iscience 24.5 (2021): 102393.
"""

import time
from pathlib import Path
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import pandas as pd


class AEModel(nn.Module):
    """GraphSCI Autoencoder class.

    Parameters
    ----------
    in_feats : int
        dimension of encoder input.
    activation : str optional
        activation function.
    n_hidden1 : int
        dimension of first hidden layer
    n_hidden2 : int
        dimension of second hidden layer
    """

    def __init__(self, in_feats, activation=None, n_hidden1=256, n_hidden2=256):
        super(AEModel, self).__init__()
        if activation != None:
            self.act1 = activation
        else:
            self.act1 = nn.Tanh()
        self.act2 = nn.ReLU()
        self.linear0 = nn.Linear(in_feats, in_feats)
        self.linear1 = nn.Linear(in_feats, n_hidden1)
        self.bn1 = nn.BatchNorm1d(n_hidden1)
        self.linear2 = nn.Linear(n_hidden1, n_hidden2)
        self.bn2 = nn.BatchNorm1d(n_hidden2)
        self.linear3 = nn.Linear(n_hidden2, in_feats)
        self.linear4 = nn.Linear(n_hidden2, in_feats)
        self.linear5 = nn.Linear(n_hidden2, in_feats)

    def forward(self, x, z_adj, size_factors):
        """Forward propagation.

        Parameters
        ----------
        x :
            input gene expression features.
        z_adj:
            adjacency matrix, usually generated from gnn model.
        size_factors:
            cell specific size factors for transforming mean of ZINB to x_exp

        Returns
        -------
        x_exp:
            recreation of input gene expression.
        pi :
            data dropout probability from ZINB.
        disp :
            data dispersion from ZINB.


        """
        x = self.linear0(x)
        x = torch.matmul(z_adj, x)
        x = self.act1(x)
        h = self.linear1(x)
        h = self.bn1(h)
        h = self.act2(h)
        h = self.linear2(h)
        h = self.bn2(h)
        h = self.act2(h)
        pi = torch.sigmoid(self.linear3(h))
        disp = torch.clamp(F.softplus(self.linear4(h)), 1e-4, 1e4)
        mean = torch.clamp(torch.exp(self.linear5(h)), 1e-5, 1e6)
        size_diag = torch.diag(size_factors).to(mean.device)
        x_exp = torch.matmul(size_diag, mean)
        return x_exp, mean, disp, pi


class GNNModel(nn.Module):
    """GraphSCI graphical neural network class.

    Parameters
    ----------
    in_feats : int
        dimension of GNN module input.
    out_feats : int
        dimension of GNN module input.
    n_hidden1 : int optional
        dimension of first hidden layer
    n_hidden2 : int optional
        dimension of second hidden layer
    activation : str optional
        activation function.
    dropout : float optional
        probability of weight dropout for training.
    """

    def __init__(self, in_feats, out_feats, n_hidden1=256, n_hidden2=256, activation=None, dropout=0.0):
        self.in_feats = in_feats
        self.out_feats = out_feats
        super(GNNModel, self).__init__()
        if dropout != 0:
            self.dropout = nn.Dropout(p=dropout)
        else:
            self.dropout = None
        self.linear1 = nn.Linear(in_feats, n_hidden1)
        self.linear2 = nn.Linear(n_hidden1, n_hidden2)
        self.linear3 = nn.Linear(n_hidden2, out_feats)
        self.linear4 = nn.Linear(n_hidden2, out_feats)
        if activation != None:
            self.act = activation
        else:
            self.act = nn.ReLU()

    def forward(self, x, adj):
        """Forward propagation.
        Parameters
        ----------
        x :
            input gene expression features.
        adj:
            gene graph adjacency matrix
        Returns
        -------
        z_adj :
            recreation of adjacency matrix
        z_adj_std :
            standard deviation parameter of normal distribution
        z_adj_mean :
            data dispersion from ZINB.
        """
        x = self.dropout(x)
        x = self.linear1(x)
        x = torch.matmul(adj, x)
        h1 = self.act(x)
        h1 = self.dropout(h1)
        h1 = self.linear2(h1)
        h1 = torch.matmul(adj, h1)
        h2 = self.act(h1)
        z_adj_mean = self.dropout(h2)
        z_adj_mean = self.linear3(z_adj_mean)
        z_adj_mean = torch.matmul(adj, z_adj_mean)
        z_adj_mean = torch.matmul(z_adj_mean, z_adj_mean.t())
        z_adj_std = self.dropout(h2)
        z_adj_std = self.linear4(z_adj_std)
        z_adj_std = torch.matmul(adj, z_adj_std)
        z_adj_std = torch.matmul(z_adj_std, z_adj_std.t())
        z_adj = z_adj_mean + torch.normal(z_adj_std, torch.ones(x.shape[0], x.shape[0]).to(z_adj_std.device))
        return z_adj, z_adj_std, z_adj_mean


class GraphSCI:
    """GraphSCI model, combination AE and GNN.

     Parameters
     ----------
     num_cells : int
         number of cells in expression data
     num_genes : int
         number of genes in expression data
     train_dataset : str
         name of training dataset
     n_epochs : int optional
         number of training epochs
     lr : float optional
         learning rate
     weight_decay : float optional
         weight decay rate
     dropout : float optional
         probability of weight dropout for training
     gpu: int optional
         index of computing device, -1 for cpu.
     """

    def __init__(self, num_cells, num_genes, train_dataset, n_epochs=100, lr=1e-3, weight_decay=1e-5, dropout=0.1,
                 gpu=-1):
        # self.params = params
        self.postfix = time.strftime('%d_%m_%Y') + '_' + time.strftime('%H:%M:%S')
        self.prj_path = Path(
            __file__).parent.resolve().parent.resolve().parent.resolve().parent.resolve().parent.resolve()
        self.train_dataset = train_dataset
        self.save_path = self.prj_path / 'example' / 'single_modality' / 'imputation' / 'pretrained' / train_dataset / 'models'
        if not self.save_path.exists():
            self.save_path.mkdir(parents=True)
        self.device = torch.device('cpu' if gpu == -1 else f'cuda:{gpu}')
        self.dropout = dropout
        self.n_epochs = n_epochs
        self.lr = lr
        self.weight_decay = weight_decay
        self.num_cells = num_cells
        self.num_genes = num_genes
        self.gnnmodel = GNNModel(in_feats=self.num_cells,
                                  out_feats=self.num_cells,
                                  dropout=self.dropout).to(self.device)
        self.aemodel = AEModel(in_feats=self.num_cells, ).to(self.device)
        self.model_params = list(self.aemodel.parameters()) + list(self.gnnmodel.parameters())
        self.optimizer = torch.optim.Adam(self.model_params, lr=self.lr, weight_decay=self.weight_decay)

    def fit(self, train_data, train_data_raw, adj_train, train_size_factors, adj_norm_train, le=1, la=1, ke=1, ka=1):
        min_train_loss, _train_loss, _epoch = 1000, 1000, 1
        """ Data fitting function 
        Parameters
        ----------
        train_data : 
            input training features
        train_data_raw : 
            input raw training features
        adj_train : 
            training adjacency matrix of gene graph
        train_size_factors : 
            train size factors for cells
        adj_norm_train : 
            normalized training adjacency matrix of gene graph
        le : float optioanl
            parameter of expression loss
        la : float optioanl
            parameter of adjacency loss
        ke : float optioanl
            parameter of KL divergence of expression
        ka : float optioanl
            parameter of KL divergence of adjacency

        Returns
        -------
        None
        """

        for epoch in range(self.n_epochs):
            total_loss = self.train(train_data, train_data_raw, adj_train, train_size_factors, adj_norm_train, le, la,
                                    ke, ka)
            train_loss, z_adj_train, z_exp_train = self.evaluate(train_data, train_data_raw, adj_norm_train, adj_train,
                                                                 train_size_factors, le, la, ke, ka)
            if min_train_loss <= train_loss:
                min_train_loss = train_loss
                _train_loss = train_loss
                _epoch = epoch
                self.save_model()
            print(f"[Epoch%d], train_loss %.6f, adj_loss %.6f, express_loss %.6f, kl loss %.6f" \
                  % (epoch, self.loss, self.loss_adj, self.loss_exp, abs(self.kl)))

    def train(self, train_data, train_data_raw, adj_orig, size_factors, adj_norm, le=1, la=1, ke=1, ka=1):
        """ Train function, gets loss and performs optimization step
        Parameters
        ----------
        train_data :
            input training features
        train_data_raw :
            input raw training features
        adj_orig :
            training adjacency matrix of gene graph
        size_factors :
            train size factors for cells
        adj_norm :
            normalized training adjacency matrix of gene graph
        le : float optioanl
            parameter of expression loss
        la : float optioanl
            parameter of adjacency loss
        ke : float optioanl
            parameter of KL divergence of expression
        ka : float optioanl
            parameter of KL divergence of adjacency


        Returns
        -------
        total_loss : float
            loss value of training loop
        """
        self.gnnmodel.train()
        self.aemodel.train()
        self.optimizer.zero_grad()
        z_adj, z_adj_std, z_adj_mean = self.gnnmodel.forward(train_data, adj_norm)
        z_exp, mean, disp, pi = self.aemodel.forward(train_data, z_adj, size_factors)
        loss_adj, loss_exp, log_lik, kl, loss = self.get_loss(train_data_raw, adj_orig, z_adj, z_adj_std, z_adj_mean,
                                                              z_exp, mean, disp, pi, size_factors, le, la, ke, ka)
        self.loss_adj = loss_adj.item()
        self.loss_exp = loss_exp.item()
        self.log_lik = log_lik.item()
        self.kl = kl.item()
        self.loss = loss.item()
        loss.backward()
        nn.utils.clip_grad_norm_(self.model_params, 1e04)
        self.optimizer.step()

        total_loss = self.loss
        del loss_adj, loss_exp, log_lik, kl, loss

        return total_loss

    def evaluate(self, features, features_raw, adj_norm, adj_orig, size_factors, le=1, la=1, ke=1, ka=1):
        """ Evaluate function, returns loss and reconstructions of expression and adjacency
        Parameters
        ----------
        features :
            input features
        features_raw :
            input raw features
        adj_norm :
            normalized adjacency matrix of gene graph
        adj_orig :
            training adjacency matrix of gene graph
        size_factors :
            cell size factors for reconstruction
        le : float optioanl
            parameter of expression loss
        la : float optioanl
            parameter of adjacency loss
        ke : float optioanl
            parameter of KL divergence of expression
        ka : float optioanl
            parameter of KL divergence of adjacency


        Returns
        -------
        """
        self.aemodel.eval()
        self.gnnmodel.eval()
        with torch.no_grad():
            z_adj, z_adj_std, z_adj_mean = self.gnnmodel.forward(features, adj_norm)
            z_exp, mean, disp, pi = self.aemodel.forward(features, z_adj, size_factors)
            loss_adj, loss_exp, log_lik, kl, loss = self.get_loss(features_raw, adj_orig, z_adj, z_adj_std, z_adj_mean,
                                                                  z_exp, mean, disp, pi, size_factors, le, la, ke, ka)
        return loss, z_adj, z_exp

    def save_model(self):
        """ Save model function, saves both AE and GNN"""
        state = {
            'aemodel': self.aemodel.state_dict(),
            'gnnmodel': self.gnnmodel.state_dict(),
            'optimizer': self.optimizer.state_dict()
        }

        torch.save(state, self.save_path / f"{self.train_dataset}.pt")

    def predict(self, data, data_raw, adj_norm, adj_orig, size_factors):
        """ Predict function
        Parameters
        ----------
        data :
            input true expression data
        data_raw :
            raw input true expression data
        adj_norm :
            normalized adjacency matrix of gene graph
        adj_orig :
            adjacency matrix of gene graph
        size_factors :
            cell size factors for reconstruction

        Returns
        -------
        z_exp :
            reconstructed expression data
        """
        _, _, z_exp = self.evaluate(data, data_raw, adj_norm, adj_orig, size_factors)
        return z_exp

    def get_loss(self, batch, adj_orig, z_adj, z_adj_std, z_adj_mean, z_exp, mean, disp, pi, sf, le=1, la=1, ke=1,
                 ka=1):
        """ Loss function for GraphSCI
        Parameters
        ----------
        batch :
            batch features
        z_adj :
            reconstructed adjacency matrix
        z_adj_std :
            standard deviation of distribution of z_adj
        z_adj_mean :
            mean of distributino of z_adj
        z_exp :
            recontruction of expression values
        mean :
            dropout parameter of ZINB dist of z_exp
        disp :
            dropout parameter of ZINB dist of z_exp
        pi :
            dispersion parameter of ZINB dist of z_exp
        sf :
            cell size factors
        le : float optioanl
            parameter of expression loss
        la : float optioanl
            parameter of adjacency loss
        ke : float optioanl
            parameter of KL divergence of expression
        ka : float optioanl
            parameter of KL divergence of adjacency

        Returns
        -------
        loss_adj : float
            loss of adjacency reconstruction
        loss_exp : float
            loss of expression reconstruction
        log_lik : float
            log likelihood loss value
        kl : float
            kullback leibler loss
        loss : float
            log_lik - kl
        """

        pos_weight = (adj_orig.shape[0] ** 2 - adj_orig.sum(axis=1)) / (adj_orig.sum(axis=1))
        norm_adj = (adj_orig.shape[0] ** 2) / (adj_orig.shape[0] ** 2 - adj_orig.sum() * 2)
        loss_adj = la * torch.mean(F.cross_entropy(z_adj, adj_orig)) / adj_orig.shape[0]

        eps = 1e-7
        t1 = torch.lgamma(disp + eps) + torch.lgamma(batch + 1) - torch.lgamma(batch + disp + eps)
        t2 = (disp + batch) * torch.log(1.0 + (mean / (disp + eps))) + (
                    batch * (torch.log(disp + eps) - torch.log(mean + eps)))
        nb_loss = t1 + t2
        nb_loss = torch.where(torch.isnan(nb_loss),
                              torch.zeros([nb_loss.shape[0], nb_loss.shape[1]]).to(self.device) + np.inf, nb_loss)
        zero_nb = torch.pow(disp / (disp + mean + eps), disp)
        zero_case = -torch.log(pi + ((1 - pi) * zero_nb) + eps)
        loss_exp = torch.where(torch.lt(batch, 1e-8), zero_case, nb_loss)
        loss_exp = le * torch.mean(loss_exp)
        log_lik = loss_exp + loss_adj

        kl_adj = (self.weight_decay * 0.5 / batch.shape[1] ** 2) * torch.mean(
            torch.sum(1 + 2 * torch.log(abs(z_adj_std)) - torch.square(z_adj_mean) - torch.square(z_adj_std), 1))
        # kl_exp = (0.5 / batch.shape[0]) * torch.mean(F.mse_loss(z_exp, batch))
        kl_exp = 0.5 * torch.mean(F.mse_loss(z_exp, batch))
        kl = ka * kl_adj - ke * kl_exp
        loss = log_lik - kl
        return loss_adj, loss_exp, log_lik, kl, loss

    def load_model(self):
        """ Load function"""
        model_path = self.prj_path / 'pretrained' / self.train_dataset / 'models' / f'{self.train_dataset}.pt'
        state = torch.load(model_path, map_location=self.device)
        self.aemodel.load_state_dict(state['aemodel'])
        self.gnnmodel.load_state_dict(state['gnnmodel'])

    def score(self, true_expr, imputed_expr, test_idx, metric="MSE"):
        """ Scoring function of model
        Parameters
        ----------
        true_expr :
            True underlying expression values
        imputed_expr :
            Imputed expression values
        test_idx :
            index of testing genes
        metric : str optional
            Choice of scoring metric - 'MSE' or 'ARI'

        Returns
        -------
        mse_error : float
            mean square error
        """
        allowd_metrics = {"MSE", "PCC"}
        true_target = true_expr[test_idx,]
        imputed_target = imputed_expr[test_idx,]
        if metric not in allowd_metrics:
            raise ValueError("scoring metric %r." % allowd_metrics)

        if (metric == 'MSE'):
            mse_cells = pd.DataFrame(((true_target.cpu() - imputed_target.cpu()) ** 2).mean(axis=0)).dropna()
            mse_genes = pd.DataFrame(((true_target.cpu() - imputed_target.cpu()) ** 2).mean(axis=1)).dropna()
            return mse_cells, mse_genes
        elif (metric == 'PCC'):
            # cor_cells = np.corrcoef(true_target.cpu(), imputed_target.cpu())
            cor_cells = np.corrcoef(true_expr.cpu(), imputed_expr.cpu())
            return cor_cells